#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

GutenbergTextParser.py

Copyright 2009 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

"""

from __future__ import unicode_literals

import re

import six
import lxml
from lxml import etree

from pkg_resources import resource_string

import libgutenberg.GutenbergGlobals as gg
from libgutenberg.GutenbergGlobals import xpath, Struct, NS
from libgutenberg.Logger import warning, info, debug
from libgutenberg.MediaTypes import mediatypes as mt

from ebookmaker import parsers
from ebookmaker.CommonCode import Options
from ebookmaker.parsers import HTMLParserBase
from ebookmaker.parsers.boilerplate import strip_headers_from_txt

options = Options()
mediatypes = (mt.txt, )

MAX_BEFORE = 5 # no. of empty lines that mark a <h1>

RE_ITALICS = re.compile(r"\b_([^_]+?)_\b")
RE_INDENT = re.compile(r"^\s+")

THRESHOLD = 1.99

# headers

HEADER_SMELLS = r"^\s*(volume|book|part|chapter|section|act|scene|table of)\b"

# always preformat

PRE_SMELLS = [
    r"gbnewby",
    r"^\s*http://",
    ]

# always reflow

P_SMELLS = [
    r"^title: ",
    r"(?:produced|prepared) by",
    r"\bebook\b",
    r"Give Away One Trillion Etext",
    r"www\.gutenberg",
    r"^\*\*\*\s*start of",
    r"^\*\*\*\s*end of",
    ]

RE_HEADER_SMELLS = re.compile(HEADER_SMELLS, re.I)
RE_P_SMELLS = re.compile("|".join(P_SMELLS), re.I)
RE_PRE_SMELLS = re.compile("|".join(PRE_SMELLS), re.I)

SUBJECTS = set('header verse quote center right'.split())

SPECIALS = {
        ord('&'): '&amp;',
        ord('<'): '&lt;',
        ord('>'): '&gt;',
        ord('"'): '&#x22;',
        0xa0:     '&#xa0;',
        }

def about_same(f1, f2):
    """ Return True if f1 and f2 are about as big. """
    if f1 is None or f2 is None:
        return False
    return max(f1, float(f2)) / min(f1, float(f2)) > 0.8

def count(iterable):
    """ Count elements that are True. """
    return float(len(list(filter(bool, iterable))))

def proportional(iterable):
    """ Return ratio True elements in iterable. """
    if iterable:
        return count(iterable) / len(iterable)
    return 0.5

def most(iterable):
    """ Return True if most of iterable is True """
    if len(iterable) < 2:
        return False
    return proportional(iterable) >= 0.75

def half(iterable):
    """ Return True if at least half of iterable is True """
    if len(iterable) < 2:
        return False
    return proportional(iterable) >= 0.5

def some(iterable):
    """ Return True if some of iterable is True """
    if len(iterable) < 2:
        return False
    return proportional(iterable) >= 0.25

def not_(iterable):
    """ Return iterable with all elements negated. """
    return [not v for v in iterable]

def and_(iterable1, iterable2):
    """ Return iterable with elements from iterables and-ed . """
    return [i[0] and i[1] for i in zip(iterable1, iterable2)]

def or_(iterable1, iterable2):
    """ Return iterable with elements from iterables or-ed . """
    return [i[0] or i[1] for i in zip(iterable1, iterable2)]


class MinMaxAvg(object):
    """ Store min, max, avg of a list of values. """

    __slots__ = "min max avg first last cnt values".split()

    def __init__(self, values):
        self.min = None
        self.max = None
        self.avg = None
        self.first = None
        self.last = None
        self.cnt = len(values)
        self.values = values

        if self.cnt:
            self.min = min(values)
            self.max = max(values)
            self.avg = sum(values) / self.cnt
            self.first = values[0]
            self.last = values[-1]


class ParagraphMetrics(object):
    """ Calculates some metrics. """

    words = None
    if hasattr(options, 'config'):
        try:
            from six.moves import dbm_gnu
            try:
                fn = options.config.RHYMING_DICT
                if fn is not None:
                    words = dbm_gnu.open(fn)
            except dbm_gnu.error:
                warning("File containing rhyming dictionary not found: %s" % fn)
        except (ModuleNotFoundError, ImportError):
            warning("No gnu dbm support found. Rhyming dictionary not used.")
    else:
        warning("No config found. Rhyming dictionary not used.")

    def __init__(self, par):
        """ Calculate metrics about this paragraph. """
        lines = par.lines

        self.cnt_lines = len(lines)

        self.lengths = list(map(len, lines))
        self.centers = list(map(self._center, lines))
        self.indents = list(map(self._indent, lines))

        self.titles = list(map(self._istitle, lines))
        self.uppers = list(map(six.text_type.isupper, lines))

        # skip last line, which is almost always shorter
        self.length = MinMaxAvg(self.lengths[:-1])
        self.length.last = self.lengths[-1]
        # skip first line, which sometimes is indented on every par
        self.indent = MinMaxAvg(self.indents[1:])
        self.indent.first = self.indents[0]
        # all lines must be centered
        self.center = MinMaxAvg(self.centers)

        self.stems = None
        self.rhymes = None
        if self.words:
            self._init_rhymes(par)


    @staticmethod
    def _indent(line):
        """ Find out how much a line is left-indented. """
        return len(line) - len(line.lstrip())

    @staticmethod
    def _center(line):
        """ Find the center pos of a line. """
        len_ = len(line)
        indent = len_ - len(line.lstrip())
        return (len_ + indent) / 2

    @staticmethod
    def _istitle(line):
        """ Return True if the first char is uppercase. """
        m = re.search(r'\w', line)
        return m and m.group(0).isupper()

    def _rhyme_stemmer(self, line):
        """ Return the stem of the rhyme.

        See comments in: rhyme_compiler.py

        """

        line = re.sub(r'\W*$', '', line)

        words = re.split('[- ]+', line)
        try:
            last_word = words[-1].lower()
            return self.words[last_word.encode('utf-8')]
        except (IndexError, KeyError):
            last_word = re.sub('^(un|in)', '', last_word)
            try:
                return self.words[last_word.encode('utf-8')]
            except (IndexError, KeyError):
                return None

    def _init_rhymes(self, par):
        """ Get rhyme stems and see which lines do rhyme. """
        self.stems = list(map(self._rhyme_stemmer, par.lines))
        self.rhymes = len(self.stems) * [0]

        go_back = 8  # how many lines to consider

        for i, stem in enumerate(self.stems):
            if stem is None:
                continue
            try:
                j = self.stems.index(stem, max(0, i - go_back), i)
                self.rhymes[j] = 1
                self.rhymes[i] = 1
            except ValueError:
                pass


class Par(object):
    """ Contains one paragraph with lots of metrics. """
#    __slots__ = ('lines styles tag before after id scores'.split())

    def __init__(self):
        self.lines = []
        self.styles = {}
        self.metrics = None
        self.tag = None
        self.before = 0
        self.after = 0
        self.id = None
        self.prev = None
        self.debug_message = ''

        self.scores = Struct()
        for subject in SUBJECTS:
            setattr(self.scores, subject, 1.0)

    def __len__(self):
        return min(len(self.lines), 50)

    def flush_left_lines(self):
        """ Return lines that are flush left.

        Note that those lines may well be indented.

        Returns array bitfield.

        """
        return [v == self.metrics.indent.min for v in self.metrics.indents]

    def centered_lines(self):
        """ Return lines that are centered. """
        return [abs(v - self.metrics.center.avg) < 2 for v in self.metrics.centers]

    def flush_right_lines(self):
        """ Return lines that are flush right.

        Note that those lines may well be very short.

        """
        return [v == self.metrics.length.max for v in self.metrics.lengths]

    def short_lines(self):
        """ Return lines much shorter than average. """
        if self.metrics.length.avg is None:
            return []
        thresh = self.metrics.length.avg / 2.0
        return [v < thresh for v in self.metrics.lengths]

    def internal_short_lines(self):
        """ Return lines much shorter than average. Except last line. """
        if self.metrics.length.avg is None:
            return []
        res = self.short_lines()
        # last line should not be considered `short´ even if it is
        res[-1] = False
        return res

    def long_lines(self):
        """ Return lines longer than average. """
        if self.metrics.length.avg is None:
            return []
        return [v > self.metrics.length.avg for v in self.metrics.lengths]

        # a sequence of pars of the same length

        # same indentation pattern as pars before and after

    def header_smells(self):
        """ Test some words we know hint at headers """
        return RE_HEADER_SMELLS.findall(" ".join(self.lines))

    def p_smells(self):
        """ Test some words we know hint at reflowed text. """
        return RE_P_SMELLS.findall(" ".join(self.lines))

    def pre_smells(self):
        """ Test some words we know hint at preformatted text. """
        return RE_PRE_SMELLS.findall(" ".join(self.lines))

    def msg(self, m):
        """ Add to debug message. """
        self.debug_message += m + ' -- '
        return m

    def fix_shorties(self):
        """ Fix any internal short lines. """

        # We also fix the last line, that may naturally be shorter on
        # paragraphs, because it doesn't matter in the case of
        # paragraphs but helps in the case of verse.

        lines = self.lines
        for i in range(1, len(lines) - 1):
            if len(lines[i]) < 25: # ad-hocked value
                if len(lines[i-1]) > 50 and lines[i-1][-1:] != '-':
                    lines[i-1] += ' ' + lines[i]
                    lines[i] = ''
        self.lines = filter(len, lines)



    def analyze(self):
        """ Guess paragraph type -- Part 1.

        Guess if this paragraph is a header, verse, quote or
        anything. Run lots of cunning tests and assign fuzzy scores.

        """

        # header ?

        if all(self.metrics.uppers):
            self.msg("all uppercase")
            self.scores.header *= 2.0

        if any(self.header_smells()):
            self.msg("any header smells")
            self.scores.header *= 2.0

        # analyze indentation

        if half(self.metrics.indents):
            self.msg("half indents")
            self.scores.quote = 2.00

        if most(or_(self.metrics.titles, self.internal_short_lines())):
            self.msg("most (titles or internal_short)")
            self.scores.quote = 2.00
            self.scores.verse *= 1.1 ** len(self)

        # verse or quote ?

        c = count(self.metrics.titles)
        self.scores.verse *= 1.2 ** (min(c - len(self.lines), 50) / 2.0)
        self.msg("%d titles in %d" % (c, len(self)))

        if self.metrics.rhymes:
            if all(self.metrics.rhymes):
                self.msg("all rhyming_lines")
                self.scores.quote *= 1.2 ** len(self)
                self.scores.verse *= 1.2 ** len(self)

            c = count(self.metrics.rhymes)
            self.scores.verse *= 1.1 ** (c - len(self) / 2.0)
            self.msg("%d rhyming_lines in %d" % (c, len(self)))

            c = count(and_(self.metrics.rhymes, self.short_lines()))
            d = count(self.short_lines())
            self.scores.verse *= 1.1 ** (c - d / 2.0)
            self.msg("%d short rhyming_lines in %d" % (c, d))

        # FIXME: inspect punctuation at end-of-line

        if some(not_(self.flush_left_lines()[1:])):
            self.msg("some (not flush_left)")
            self.scores.verse *= 20.0 # strong indicator

        if any(self.internal_short_lines()):
            self.msg("any internal_short_lines")
            self.scores.verse *= 20.0 # strong indicator

        if any(self.p_smells()):
            self.msg("any p smells")
            self.scores.header = 0.0
            self.scores.quote = 0.0

        if any(self.pre_smells()):
            self.msg("any pre smells")
            self.scores.header = 0.0
            self.scores.quote = 2.0
            self.scores.verse = 2.0


    def analyze_multi(self):
        """ Guess paragraph type -- Part 2.

        Tests spanning multiple paragraphs.

        """

        if self.prev:
            if (any(not_(self.flush_left_lines())) and
                    self.prev.metrics.indents == self.metrics.indents):
                # same indentation scheme (implies same line count)
                self.msg(self.prev.msg("same indentation as neighbor"))
                self.scores.verse *= 2.0
                self.prev.scores.verse *= 2.0

            if self.prev.scores.quote > THRESHOLD and self.prev.scores.verse > 1.0:
                self.msg("follows verse")
                self.scores.quote *= 1.2
                self.scores.verse *= 1.2

            if self.scores.quote > THRESHOLD and self.scores.verse > 1.0:
                self.prev.msg("precedes verse")
                self.prev.scores.quote *= 1.2
                self.prev.scores.verse *= 1.2

            if (self.metrics.cnt_lines == self.prev.metrics.cnt_lines and
                    about_same(self.metrics.length.avg, self.prev.metrics.length.avg)):
                self.msg(self.prev.msg("same look as neighbor"))
                self.scores.verse *= 1.2
                self.prev.scores.verse *= 1.2



class Parser(HTMLParserBase):
    """Parse a Project Gutenberg 'Plain Vanilla Text'

    and convert to xhtml suitable for ePub packaging.

    """

    def __init__(self, attribs=None):
        HTMLParserBase.__init__(self, attribs)
        self.body = 0
        self.max_blanks = 0
        self.pars = []


    def get_charset_from_meta(self):
        """ Parse text for hints about charset. """

        charset = None

        match = parsers.REB_PG_CHARSET.search(self.bytes_content())
        if match:
            charset = match.group(1).decode('ascii')
            info('Got charset %s from pg header' % charset)

        return charset


    def analyze(self):
        """ analyze parsed paragraphs

        do all sorts of smart stuff here

        """

        last_par = None
        for par in self.pars:
            # par.fix_shorties()
            par.metrics = ParagraphMetrics(par)
            par.prev = last_par
            if last_par:
                last_par.next = par
            last_par = par

        for par in self.pars:
            par.analyze()

        # second run for analyses spanning multiple paragraphs
        # may use results from first run
        for par in self.pars:
            par.analyze_multi()

        for par in self.pars:
            par.msg("header: %f" % par.scores.header)
            par.msg("verse: %f" % par.scores.verse)
            par.msg("quote: %f" % par.scores.quote)
            par.msg("center: %f" % par.scores.center)
            par.msg("right: %f" % par.scores.right)


        # translate findings into css styles

        for n, par in enumerate(self.pars):
            par.tag = 'p'
            par.id = "id%05d" % n

            if par.before > 1:
                par.styles['margin-top'] = "%dem" % par.before

            if par.scores.header > THRESHOLD:
                level = max(MAX_BEFORE - par.before, 0)
                par.tag = "h%d"  % (level + 1)
            else:
                if par.scores.quote > THRESHOLD:
                    if par.scores.verse > 1.0:
                        par.styles['white-space'] = 'pre'
                    else:
                        par.styles['margin-left'] = '%d%%' % (
                            par.metrics.indent.first * 100 / 72)
                        par.styles['margin-right'] = par.styles['margin-left']

                    if par.scores.right > THRESHOLD:
                        par.styles['text-align'] = 'right'
                    if par.scores.center > THRESHOLD:
                        par.styles['text-align'] = 'center'

    @staticmethod
    def preformat(line):
        """ Format paragraph as pre. """
        m = RE_INDENT.match(line)
        if m:
            # 0x0a   no-break space
            # 0x2003 em-space
            # 0x2007 figure space
            line = ('&#xa0;' * (m.end() - m.start())) + line[m.end():]
        return line + "<br />\n"


    def ship_out(self, par):
        """ ready paragraph for shipping """
        def italics(s):
            """ replace underscores with <i>...</i> """
            def it_repl(matchobj):
                """ helper """
                return '<i>%s</i>' % matchobj.group(1)

            return RE_ITALICS.sub(it_repl, s)

        if par.styles.get('white-space', '') == 'pre':
            par.lines = map(self.preformat, par.lines)
            del par.styles['white-space']

        text = italics("\n".join(par.lines))
        text = text.replace("--", "&#x2014;")
        text = text.replace("...", "&#x2026;")

        style = ''
        if par.styles:
            styles = []
            for s, v in par.styles.items():
                styles.append("%s: %s" % (s, v))
            style = ' style="' + "; ".join(styles) + '"'

        id_ = ''
        if par.id:
            id_ = ' id="%s"' % par.id

        title = ''
        if options.verbose >= 3 and par.debug_message:
            title = ' title="%s"' % par.debug_message
        # title = ' title="%s"' % repr(most(not_(par.metrics.titles)))

        ns = ' xmlns="%s"' % str(NS.xhtml)
        return '<%s%s%s%s%s>%s</%s>' % (par.tag, ns, id_, style, title, text, par.tag)


    def iterlinks(self): # pylint: disable=R0201
        """ There are no links in text files. """
        return []


    def rewrite_links(self, f): # pylint: disable=R0201
        """ There are no links in text files. """
        return


    def pre_parse(self):
        """ Nothing to do here, because there are no links in text
        files.  iterlinks() will simply return an empty list."""

        debug("GutenbergTextParser.pre_parse() ...")


    def css_content(self):
        default_css = resource_string(
            'ebookmaker.parsers', 'txt2all.css').decode('utf-8')
        return default_css


    def parse(self):
        """ Parse the plain text.

        Try to find semantic units in the character soup. """

        debug("GutenbergTextParser.parse() ...")

        if self.xhtml is not None:
            return

        text = self.unicode_content()
        text, pg_header, pg_footer = strip_headers_from_txt(text)
        text = parsers.RE_RESTRICTED.sub('', text)
        text = gg.xmlspecialchars(text)

        lines = [line.rstrip() for line in text.splitlines()]
        lines.append("")
        del text

        blanks = 0
        par = Par()

        for line in lines:
            if len(line) == 0:
                blanks += 1
            else:
                if blanks and par.lines: # don't append empty pars
                    par.after = blanks
                    self.pars.append(par)
                    if self.body == 1:
                        self.max_blanks = max(blanks, self.max_blanks)
                    par = Par()
                    par.before = blanks
                    blanks = 0

                par.lines.append(line)

        par.after = blanks
        if par.lines:
            self.pars.append(par)

        lines = None

        self.analyze()

        # build xhtml tree

        em = parsers.em
        self.xhtml = em.html(
            em.head(
                em.title(' '),
                em.meta(**{'http-equiv': 'Content-Style-Type',
                           'content': 'text/css'}),
                em.meta(**{'http-equiv': 'Content-Type',
                           'content': mt.xhtml + '; charset=utf-8'}),
                em.style(self.css_content(), **{'type': 'text/css'})
            ),
            em.body()
        )

        for body in xpath(self.xhtml, '//xhtml:body'):
            xhtmlparser = lxml.html.XHTMLParser()
            if pg_header:
                body.append(etree.fromstring(pg_header, xhtmlparser))
            for par in self.pars:
                p = etree.fromstring(self.ship_out(par), xhtmlparser)
                p.tail = '\n\n'
                body.append(p)
            if pg_footer:
                body.append(etree.fromstring(pg_footer, xhtmlparser))

        self.pars = []

    def _make_coverpage_link(self, coverpage_url=None):
        """ Insert a <link rel="coverpage"> in the html head
        using the image specified by the --cover command-line option
        """

        if coverpage_url:
            for head in xpath(self.xhtml, "/xhtml:html/xhtml:head"):
                head.append(parsers.em.link(rel='icon', href=coverpage_url, format='image/x-cover'))
                debug("Inserted link to coverpage %s." % coverpage_url)
            return

    def add_title(self, dc):
        if dc.title:
            for elem in xpath(self.xhtml, '//xhtml:title'):
                dc.title = re.sub(r'\s*[\r\n]+\s*', '\n', dc.title)
                elem.text = f'The Project Gutenberg eBook of {dc.title}, by {dc.authors_short()}'
                break


#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

HTMLWriter.py

Copyright 2009 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

"""


import copy
import os
from pathlib import Path
import re
from urllib.parse import urlparse, urljoin
import uuid

from lxml import etree


from libgutenberg.Logger import debug, exception, info, error, warning

from ebookmaker import writers
from ebookmaker.CommonCode import Options
from ebookmaker.parsers import webify_url
from ebookmaker.parsers.CSSParser import cssutils
from ebookmaker.utils import (
    add_class, add_style, css_len, check_lang, replace_elements, gg, xpath, NS
)
from ebookmaker.writers import HtmlTemplates
options = Options()
cssutils.ser.prefs.validOnly = True


CSS_FOR_REPLACED = {
    'big': ".xhtml_big {font-size: larger;}",
    'tt': ".xhtml_tt {font-family: monospace;}",
    # add some needed CSS3
    '*': ".xhtml_center {justify-content: center; display: flex;}",
}

## from https://hg.mozilla.org/mozilla-central/file/3fd770ef6a65/layout/style/html.css#l310

CSS_FOR_RULES = {
    'none': '''
.rules-none > tr > td, .rules-none > * > tr > td, .rules-none > tr > th,
.rules-none > * > tr > th, .rules-none > td, .rules-none > th {
border-width: thin;
border-style: none;}''',
    'all': '''
.rules-all > tr > td, .rules-all > * > tr > td, .rules-all > tr > th,
.rules-all > * > tr > th, .rules-all > td, .rules-all > th {
border-width: thin;
border-style: solid;}''',
    'cols': '''
.rules-cols > tr > td, .rules-cols > * > tr > td,
.rules-cols > tr > th, .rules-cols > * > tr > th {
border-left-width: thin;
border-right-width: thin;
border-left-style: solid;
border-right-style: solid;}''',
    'rows': '''
.rules-rows > tr, .rules-rows > * > tr {
border-top-width: thin;
border-bottom-width: thin;
border-top-style: solid;
border-bottom-style: solid;}''',
    'groups': '''
.rules-groups > tfoot, .rules-groups > thead, .rules-groups > tbody {
border-top-width: thin; border-bottom-width: thin;
border-top-style: solid; border-bottom-style: solid;}
.rules-groups > colgroup {
border-left-width: thin; border-right-width: thin;
border-left-style: solid; border-right-style: solid;}''',
}

CSS_FOR_FRAME = {
    'box': '.frame-box {border: thin outset;}',
    'void': '.frame-void {border-style: hidden;}',
    'above': 'frame-above {border-style: outset hidden hidden hidden;}',
    'below': 'frame-below {border-style: hidden hidden outset hidden;}',
    'lhs': '.frame-lhs {border-style: hidden hidden hidden outset;}',
    'rhs': '.frame-rhs {border-style: hidden outset hidden hidden;}',
    'hsides': '.frame-hsides {border-style: outset hidden;}',
    'vsides': '.frame-vsides {border-style: hidden outset;}',
    'border': 'frame-border {border-style: outset;}',
}
CSS_FOR_REPLACED.update(CSS_FOR_RULES)
CSS_FOR_REPLACED.update(CSS_FOR_FRAME)

REPLACE_ELEMENTS = {
    'big': 'span',
    'tt': 'span',
}

# replacing attributes with css in a style attribute
# (tag, attr, cssprop, val2css)
REPLACEMENTS = [
    ('tbody td tfoot th thead tr', 'align', 'text-align', lambda x: x),
    ('table td', 'background', '', ''),
    ('table', 'border', 'border-width', css_len),
    ('table td tr', 'bordercolor', 'border-color', lambda x: x),
    ('table', 'cellpadding', 'padding', css_len),
    ('table', 'cellspacing', 'border-spacing', css_len),
    ('col td th tr', 'valign', 'vertical-align', lambda x: x),
    ('tbody tfoot thead', 'valign', 'vertical-align', lambda x: x),
    ('col', 'width', 'width', css_len),
]

DIVIDER = re.compile(r'\*\*+.*\*\*+')

def serialize(xhtml):
    """ mode is html or xml """
    return etree.tostring(xhtml,
                          method='html',
                          doctype=gg.HTML5_DOCTYPE,
                          encoding='utf-8',
                          pretty_print=False)

class Writer(writers.HTMLishWriter):
    """ Class for writing HTML files. """
    VALIDATOR = 'HTML_VALIDATOR'

    def add_dublincore(self, job, tree):
        """ Add dublin core metadata to <head>. """
        source = gg.archive2files(
            options.ebook, job.url)

        if hasattr(options.config, 'FILESDIR'):
            job.dc.source = source.replace(options.config.FILESDIR, options.config.PGURL)

        for head in xpath(tree, '//xhtml:head'):
            for e in job.dc.to_html():
                e.tail = '\n'
                head.append(e)

    def add_moremeta(self, job, tree, url):

        self.add_prop(tree, "og:title", job.dc.title)

        for dcmitype in job.dc.dcmitypes:
            self.add_prop(tree, "og:type", dcmitype.id)
        info(job.main)
        web_url = urljoin(job.dc.canonical_url, job.outputfile)
        self.add_prop(tree, "og:url", web_url)
        canonical_cover_name = 'pg%s.cover.medium.jpg' % job.dc.project_gutenberg_id
        cover_url = urljoin(job.dc.canonical_url, canonical_cover_name)
        self.add_prop(tree, "og:image", cover_url)

        # fix empty title elements
        for title in xpath(tree, '//xhtml:title[not(text())]'):
            title.text = job.dc.title

    @staticmethod
    def replace_boilerplate(job, tree):
        # get text after the divider, put in dc.credit if not already there
        if not job.dc.credit:
            for pre in xpath(tree, '//*[@id="pg-header"]//xhtml:pre'):
                divided = DIVIDER.split(' '.join(pre.itertext()))
                if len(divided) > 1:
                    job.dc.credit = divided[1]

        body = None
        for body in xpath(tree, '//xhtml:body'):
            break
        else:
            error('No Body!!!')
            return
        
        new_bp = HtmlTemplates.pgheader(job.dc)

        for pg_header in xpath(tree, '//*[@id="pg-header"]'):
            prev = pg_header.getprevious()
            parent = pg_header.getparent()
            while prev is not None:
                parent.remove(prev)
                prev = pg_header.getprevious()
            parent.replace(pg_header, new_bp)
            break
        else:
            body.insert(0, new_bp)
            warning('No pg-header found, inserted a generated one')
            
        new_bp = HtmlTemplates.pgfooter(job.dc)

        for pg_footer in xpath(tree, '//*[@id="pg-footer"]'):
            next = pg_footer.getnext()
            parent = pg_footer.getparent()
            while next is not None:
                parent.remove(next)
                next = pg_footer.getnext()
            parent.replace(pg_footer, new_bp)
            break
        else:
            body.append(new_bp)
            warning('No pg-footer found, inserted a generated one')

        for pg_smallprint in xpath(tree, '//*[@id="pg-smallprint"]'):
            pg_smallprint.getparent().remove(pg_smallprint)


    def outputfileurl(self, job, url):
        """
        Make the output path for the parser.
        Consider an image referenced in a source html file being moved to a destination directory.
        The image must be moved to a Location that is the same, relative to the job's destination,
        as it was in the source file.
        The constraints are that
        1. we must not over-write the source files, and
        2. the destination directory may be the same as the source directory.
        In case (2), we'll create a new "out" directory to contain the written files; we'll also
        stop with an error if our source path is below an "out" directory.

        Complication: generated covers are already in the output directory.

        """

        if not job.main:
            # this is the main file.
            job.main = url

            # check that the source file is not in the outputdir
            if gg.is_same_path(os.path.abspath(job.outputdir), os.path.dirname(url)):
                # make sure that source is not in an 'out" directory
                newdir = 'out'
                for parent in Path(url).parents:
                    if parent.name == newdir:
                        # not allowed
                        newdir = uuid.uuid4().hex
                        warning("can't use an 'out' directory for both input and output; using %s",
                                newdir)
                        break

                job.outputdir = os.path.join(job.outputdir, newdir)

            jobfilename = os.path.join(os.path.abspath(job.outputdir), job.outputfile)

            info("Creating HTML file: %s" % jobfilename)

            relativeURL = os.path.basename(url)
            if relativeURL != job.outputfile:
                # need to change the name for main file
                debug('changing %s to   %s', relativeURL, job.outputfile)
                job.link_map[relativeURL] = jobfilename
                relativeURL = job.outputfile

        else:
            if url.startswith(webify_url(job.outputdir)):
                relativeURL = gg.make_url_relative(webify_url(job.outputdir) + '/', url)
                debug('output relativeURL for %s to %s : %s', webify_url(job.outputdir), url, relativeURL)
            else:
                relativeURL = gg.make_url_relative(job.main, url)
                debug('relativeURL for %s to %s : %s', job.main, url, relativeURL)

        return os.path.join(os.path.abspath(job.outputdir), relativeURL)


    @staticmethod
    def fix_incompatible_css(sheet):
        """ Strip CSS properties and values that are not HTML5 compatible.
            Unpack "media handheld" rules
        """
        for rule in sheet:
            if rule.type == rule.STYLE_RULE:
                for p in list(rule.style):
                    if p.name == 'speak':
                        val = rule.style.getPropertyValue('speak')
                        if val == 'spell-out':
                            rule.style.removeProperty('speak')
                            rule.style.setProperty('speak-as', 'spell-out')
                        elif val == 'none':
                            rule.style.setProperty('speak', 'never')
                        elif val == 'inherit':
                            rule.style.setProperty('speak', 'auto')

    @staticmethod
    def fix_css_for_deprecated(sheet, tags='', replacement='span'):
        """ for deprecated tags, change selector to {replacement}.xhtml_{tag name};
            if no existing selector, add the selector with a style
        """
        for tag in tags:
            tagre = re.compile(f'(^| |\\+|,|>|~){tag}')
            tagsub = f'\\1{replacement}.xhtml_{tag}'
            for rule in sheet:
                if rule.type == rule.STYLE_RULE:
                    for selector in rule.selectorList:
                        selector.selectorText = tagre.sub(tagsub, selector.selectorText)

    @staticmethod
    def xhtml_to_html(html):
        '''
        try to convert the html4 DOM to an html5 DOM
        (assumes xhtml namespaces have been removed, except from attribute values)
        '''

        # fix metas
        for meta in xpath(html, "//xhtml:meta[translate(@http-equiv, 'CT', 'ct')='content-type']"):
            meta.getparent().remove(meta)
        for meta in xpath(html, "//xhtml:meta[translate(@http-equiv, 'CST', 'cst')='content-style-type']"):
            meta.getparent().remove(meta)
        for meta in xpath(html, "//xhtml:meta[translate(@http-equiv, 'CL', 'cl')='content-language']"):
            meta.getparent().remove(meta)
        for meta in xpath(html, "//xhtml:meta[@charset]"): # html5 doc, we'll replace it
            meta.getparent().remove(meta)
        for meta in xpath(html, "//xhtml:meta[@scheme]"): # remove obsolete formatted metas
            meta.getparent().remove(meta)
        for elem in xpath(html, "//xhtml:*[@xml:space]"):
            if elem.tag in ('pre', 'style'):
                del elem.attrib[NS.xml.space]

        #check values of lang; lang attributes moved to xml:lang in pre_parse
        for elem in xpath(html, "//xhtml:*[@lang]"):
            check_lang(elem, 'lang')
        for elem in xpath(html, "//xhtml:*[@xml:lang and not(@lang)]"):
            check_lang(elem, NS.xml.lang)

        # remove obsolete attributes
        attrs_to_remove = [('style', 'type'), ('img', 'longdesc')]
        for (tag, attr) in attrs_to_remove:
            for elem in xpath(html, f"//xhtml:{tag}[@{attr}]"):
                del elem.attrib[attr]

        # set required attributes
        attrs_to_fill = [('img', 'alt', '')]
        for (tag, attr, fill) in attrs_to_fill:
            for elem in xpath(html, f"//xhtml:{tag}[not(@{attr})]"):
                elem.set(attr, fill)

        # remove not_empty attributes
        nullattrs_to_remove = ['height', 'width']
        for attr in nullattrs_to_remove:
            for elem in xpath(html, f"//xhtml:*[@{attr}='' or @{attr}=0]"):
                del elem.attrib[attr]

        # put a space in empty headings
        for heading in [f'h{n}' for n in range(1,6)]:
            for elem in xpath(html, f"//xhtml:{heading}[not(text())]"):
                elem.text = ' '

        for (tags, attr, cssattr, val2css) in REPLACEMENTS:
            for tag in tags.split():
                for elem in xpath(html, f"//xhtml:{tag}[@{attr}]"):
                    if elem.attrib[attr]:
                        val = elem.attrib[attr]
                        del elem.attrib[attr]
                        if cssattr:
                            add_style(elem, style=f'{cssattr}: {val2css(val)};')


        # width and height attributes must be integer
        for elem in xpath(html, "//xhtml:*[@width or @height]"):
            rules = []
            for key in ['width', 'height']:
                if key in elem.attrib and elem.attrib[key]:
                    val = elem.attrib[key]
                    try:
                        val = int(val)
                    except ValueError:
                        del elem.attrib[key]
                        rules.append('%s: %s' % (key, val))
            if rules:
                elem.attrib['style'] = '; '.join(rules) + '; ' + elem.attrib.get('style', '')

        # fix missing <dd>,<dt> elements
        for dt in xpath(html, "//xhtml:dt"):
            if dt.getnext() is None or dt.getnext().tag != NS.xhtml.dd:
                dt.addnext(etree.Element(NS.xhtml.dd))
        for dd in xpath(html, "//xhtml:dd"):
            if dd.getprevious() is None:
                dd.addprevious(etree.Element(NS.xhtml.dt))

        # deprecated elements -  replace with <span/div class="xhtml_{tag name}">
        deprecated_used = replace_elements(html, REPLACE_ELEMENTS)


        html.head.insert(0, etree.Element(NS.xhtml.meta, charset="utf-8"))

        ##### tables #######

        # remove summary attribute
        for table in xpath(html, '//xhtml:table[@summary]'):
            summary = table.attrib['summary']
            del table.attrib['summary']
            if summary:
                table.attrib['data-summary'] = summary


        # replace frame and rules attributes on tables

        deprecated_atts = {'frame': CSS_FOR_FRAME, 'rules': CSS_FOR_RULES, 'background': {}}
        for att in deprecated_atts:
            for table in xpath(html, f'//xhtml:table[@{att}]'):
                att_value = table.attrib[att].lower()
                if att_value in deprecated_atts[att]:
                    add_class(table, f'{att}-{att_value}')
                    deprecated_used.add(att_value)
                del table.attrib[att]

        # remove span attribute from colgroups that have col children
        for colgroup in xpath(html, "//xhtml:colgroup[@span and xhtml:col]"):
            del colgroup.attrib['span']

        # move tfoot elements to end of table
        for tfoot in xpath(html, "//xhtml:table/xhtml:tfoot"):
            table = tfoot.getparent()
            table.append(tfoot)


        ##### cleanup #######

        # fix css in style elements
        cssparser = cssutils.CSSParser()
        for style in xpath(html, "//xhtml:style"):
            if style.text:
                sheet = cssparser.parseString(style.text)
                Writer.fix_incompatible_css(sheet)
                Writer.fix_css_for_deprecated(sheet, tags=deprecated_used)
                style.text = sheet.cssText.decode("utf-8")

        deprecated_used.add('*')  # '*' provides for css rules that are always added
        css_for_deprecated = ' '.join([CSS_FOR_REPLACED.get(tag, '') for tag in deprecated_used])
        if css_for_deprecated.strip():
            elem = etree.Element(NS.xhtml.style)
            elem.text = css_for_deprecated
            html.find(NS.xhtml.head).insert(1, elem) # right after charset declaration


    def build(self, job):
        """ Build HTML file. """

        def rewrite_links(job, node):
            """ only needed if the mainsource filename has been changed """
            for renamed_path in job.link_map:
                for link in xpath(node, '//xhtml:*[@href]'):
                    old_link = link.get('href')
                    parsed_url = urlparse(old_link)
                    if os.path.basename(parsed_url.path) == renamed_path:
                        new_path = parsed_url.path[0:-len(renamed_path)]
                        new_link = job.link_map[renamed_path]
                        new_link = '%s%s#%s' % (new_path, new_link, parsed_url.fragment)
                        link.set('href', new_link)

        for p in job.spider.parsers:
            # Do html only. The images were copied earlier by PicsDirWriter.

            outfile = self.outputfileurl(job, p.attribs.url)
            outfile = gg.normalize_path(outfile)

            if gg.is_same_path(p.attribs.url, outfile):
                # debug('%s is same as %s: won't override source', p.attribs.url, outfile)
                continue

            gg.mkdir_for_filename(outfile)

            xhtml = None
            if hasattr(p, 'rst2html'):
                xhtml = p.rst2html(job)
                self.make_links_relative(xhtml, p.attribs.url)
                rewrite_links(job, xhtml)

            elif hasattr(p, 'xhtml'):
                p.parse()
                xhtml = copy.deepcopy(p.xhtml)
                self.make_links_relative(xhtml, p.attribs.url)
                rewrite_links(job, xhtml)

            else:
                p.parse()

            try:

                if xhtml is not None:
                    html = copy.deepcopy(xhtml)

                    if NS.xml.lang in html.attrib:
                        lang = html.attrib[NS.xml.lang]
                        html.attrib['lang'] = job.dc.languages[0].id or lang
                        del html.attrib[NS.xml.lang]

                    self.add_dublincore(job, html)
                    self.add_meta_generator(html)
                    self.add_moremeta(job, html, p.attribs.url)

                    self.replace_boilerplate(job, html)
                    self.xhtml_to_html(html)

                    # strip xhtml namespace
                    # https://stackoverflow.com/questions/18159221/
                    for elem in html.getiterator():
                        if elem.tag is not etree.Comment:
                            elem.tag = etree.QName(elem).localname
                    # Remove unused namespace declarations
                    etree.cleanup_namespaces(html)

                    self.write_with_crlf(outfile, serialize(html))                   
                    info("Done generating HTML file: %s" % outfile)

                else:
                    #images and css files

                    if hasattr(p, 'sheet') and p.sheet:
                        self.fix_incompatible_css(p.sheet)

                    try:
                        with open(outfile, 'wb') as fp_dest:
                            fp_dest.write(p.serialize())
                    except IOError as what:
                        error('Cannot copy %s to %s: %s', job.attribs.url, outfile, what)

            except Exception as what:
                exception("Error building HTML %s: %s" % (outfile, what))
                if os.access(outfile, os.W_OK):
                    os.remove(outfile)
                raise what

        info("Done generating HTML: %s" % job.outputfile)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""

epub2.py

Copyright 2012 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

A writer that writes XHTML 1 files suited for conversion into EPUB2.

"""

import re

# from libgutenberg.Logger import info, debug, warning, error

from ebookmaker.mydocutils.writers.xhtml1 import Writer as WriterBase
from ebookmaker.mydocutils.writers.xhtml1 import Translator as TranslatorBase


class Writer (WriterBase):
    """ EPUB2 writer. """

    def __init__ (self):
        WriterBase.__init__ (self)
        self.translator_class = Translator


class Translator (TranslatorBase):
    """ HTML Translator with EPUB2 tweaks. """

    def init_css (self):
        for css_file in ('rst2all.css', 'rst2epub.css'):
            self.head.append ('<style type="text/css">\n%s</style>\n' %
                              self.encode (self.read_css (css_file)))


    def calc_centering_style (self, node):
        """
        Rationale: The EPUB standard allows user agents to replace
        `margin: auto` with `margin: 0`. Thus we cannot use `margin: auto`
        to center images, we have to calculate the left margin value.

        Also we must use 'width' on the html element, not css style,
        or Adobe ADE will not scale the image properly (ie. only
        horizontally).

        :align: is supposed to work on blocks. It floats or centers
        a block.

        :align: center has not the same semantics as :class: center.
        Former centers the block, eg. the whole table, latter centers
        the text, eg, the text in every table cell.

            `:align: center`
                Used on image: centers image
                Used on figure: centers image and caption
                Used on table: centers table and caption

        """

        width = node.get ('width')
        if width is None:
            return []

        style = ['width: %s' % width]

        m = re.match (r'(\d+)\s*%', width)
        if m:
            width = max (min (int (m.group (1)), 100), 0)
            margin = 100 - width

            align = node.get ('align', 'center')
            if align == 'center':
                style.append ('margin-left: %d%%' % (margin / 2))
            if align == 'right':
                style.append ('margin-left: %d%%' % margin)

        node['styles'].extend (style)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""

EpubWriter.py

Copyright 2009-2012 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Writes an EPUB file.

"""

import re
import datetime
import zipfile
import time
import os
import copy

from xml.sax.saxutils import quoteattr

from six.moves import urllib
from lxml import etree
from lxml.builder import ElementMaker
from pkg_resources import resource_string # pylint: disable=E0611

from libgutenberg.GutenbergGlobals import NS
from libgutenberg.Logger import critical, debug, error, exception, info, warning
from libgutenberg.MediaTypes import mediatypes as mt
from ebookmaker import parsers
from ebookmaker import ParserFactory
from ebookmaker import HTMLChunker
# from ebookmaker import Spider
from ebookmaker import writers
from ebookmaker.CommonCode import Options
from ebookmaker.Version import VERSION, GENERATOR
from ebookmaker.utils import (
    add_class, add_style, css_len, gg, xpath,
)

options = Options()

MAX_CHUNK_SIZE = 300 * 1024  # bytes
MAX_NOIMAGE_SIZE = 64 * 1024 - 1 # bytes, used for covers and important images in "noimages epubs"
MAX_IMAGE_SIZE = 256 * 1024 - 1  # in bytes, used for Kindle, too.
LINKED_IMAGE_SIZE = 1024 * 1024 - 1  # in bytes

MAX_IMAGE_DIMEN = (5000, 5000)  # in pixels
LINKED_IMAGE_DIMEN = (5000, 5000)  # in pixels
MAX_COVER_DIMEN = MAX_IMAGE_DIMEN

# iPhone 3G:    320x480x?
# Kindle 2:     600x800x16
# Sony PRS-505: 584x754x8   (effective according to wikipedia)


EPUB_TYPE = '{%s}type' % NS.epub

TOC_HEADERS = ('contents', 'table of contents',
               'inhalt', 'inhaltsverzeichnis',
               'table des matières',
               'indice',
               'inhoud')

DP_PAGENUMBER_CLASSES = frozenset('pagenum pageno page pb folionum foliono'.split())

STRIP_CLASSES = DP_PAGENUMBER_CLASSES | frozenset('versenum verseno'.split())


# Undo the more bone-headed PG and DP formattings. Set small margins
# to save on precious mobile screen real-estate.

PRIVATE_CSS = """\
@charset "utf-8";

body, body.tei.tei-text {
   color: black;
   background-color: white;
   margin: 0.5em;
   width: auto;
   border: 0;
   padding: 0;
   }
div, p, pre, h1, h2, h3, h4, h5, h6 {
   margin-left: 0;
   margin-right: 0;
   display: block
   }
div.pgebub-root-div {
   margin: 0
   }
h2 {
   page-break-before: always;
   padding-top: 1em
   }
div.figcenter span.caption {
   display: block;
   }
.pgmonospaced {
   font-family: monospace;
   font-size: 0.9em
   }
a.pgkilled {
   text-decoration: none;
   }
img.ebookmaker-cover {max-width: 100%;}
"""

OPS_TEXT_MEDIATYPES = set((
    'application/xhtml+xml',       # Used for OPS Content Documents
    'application/x-dtbook+xml',    # Used for OPS Content Documents
    'text/css',                    # Used for OPS CSS-subset style sheets
    'application/xml',             # Used for Out-Of-Line XML Islands
    'text/x-oeb1-document',        # Deprecated
    'text/x-oeb1-css',             # Deprecated
    'application/x-dtbncx+xml',    # The NCX
))

OPS_IMAGE_MEDIATYPES = set((
    'image/gif',                   # Used for raster graphics
    'image/jpeg',                  # Used for raster graphics
    'image/png',                   # Used for raster graphics
    'image/svg+xml',               # Used for vector graphics
))

OPS_CORE_MEDIATYPES = OPS_TEXT_MEDIATYPES | OPS_IMAGE_MEDIATYPES

OPS_CONTENT_DOCUMENTS = set((
    'application/xhtml+xml',
    'application/x-dtbook+xml',
    'text/x-oeb1-document',        # Deprecated
    'application/xml',
))

OPS_FONT_TYPES = set(('application/font-woff', 'application/vnd.ms-opentype'))

match_link_url = re.compile(r'^(https?://|mailto:)', re.I)
match_non_link = re.compile(r'[a-zA-Z0-9_\-\.]*(#.*)?$')

class OEBPSContainer(zipfile.ZipFile):
    """ Class representing an OEBPS Container. """

    def __init__(self, filename, oebps_path=None):
        """ Create the zip file.

        And populate it with mimetype and container.xml files.

        """

        self.zipfilename = filename
        self.oebps_path = oebps_path if oebps_path else 'OEBPS/'
        info('Creating Epub file: %s' % filename)

        # open zipfile
        zipfile.ZipFile.__init__(self, filename, 'w', zipfile.ZIP_DEFLATED)

        # write mimetype
        # the OCF spec says mimetype must be first and uncompressed
        i = self.zi()
        i.compress_type = zipfile.ZIP_STORED
        i.filename = 'mimetype'
        self.writestr(i, 'application/epub+zip')

        self.add_container_xml('content.opf')

        self.wrappers = 0 # to generate unique filenames for wrappers


    def commit(self):
        """ Close OCF Container. """
        info("Done Epub file: %s" % self.zipfilename)
        self.close()


    def rollback(self):
        """ Remove OCF Container. """
        debug("Removing Epub file: %s" % self.zipfilename)
        os.remove(self.zipfilename)


    def add_unicode(self, name, u):
        """ Add file to zip from unicode string. """
        i = self.zi(name)
        self.writestr(i, u.encode('utf-8'))


    def add_bytes(self, name, bytes_, mediatype=None):
        """ Add file to zip from bytes string. """

        i = self.zi(name)
        if mediatype and mediatype in parsers.ImageParser.mediatypes:
            i.compress_type = zipfile.ZIP_STORED
        self.writestr(i, bytes_)


    def add_file(self, name, url, mediatype=None):
        """ Add file to zip from bytes string. """

        with open(url) as fp:
            self.add_bytes(name, fp.read(), mediatype)


    def zi(self, filename=None):
        """ Make a ZipInfo. """
        z = zipfile.ZipInfo()
        z.date_time = time.gmtime()
        z.compress_type = zipfile.ZIP_DEFLATED
        z.external_attr = 0x81a40000
        if filename:
            z.filename = os.path.join(self.oebps_path, filename)
        return z


    def add_container_xml(self, rootfilename):
        """ Write container.xml

        <?xml version='1.0' encoding='UTF-8'?>

        <container xmlns='urn:oasis:names:tc:opendocument:xmlns:container'
                   version='1.0'>
          <rootfiles>
            <rootfile full-path='$path'
                      media-type='application/oebps-package+xml' />
          </rootfiles>
        </container>

        """

        rootfilename = os.path.join(self.oebps_path, rootfilename)

        ns_oasis = 'urn:oasis:names:tc:opendocument:xmlns:container'

        ocf = ElementMaker(namespace=ns_oasis,
                           nsmap={None: ns_oasis})

        container = ocf.container(
            ocf.rootfiles(
                ocf.rootfile(**{
                    'full-path': rootfilename,
                    'media-type': 'application/oebps-package+xml'})),
            version='1.0')

        i = self.zi()
        i.filename = 'META-INF/container.xml'
        self.writestr(i, etree.tostring(
            container, encoding='utf-8', xml_declaration=True, pretty_print=True))


    def add_image_wrapper(self, img_url, img_title):
        """ Add a HTML file wrapping img_url. """
        img_title = quoteattr(img_title)
        filename = 'wrap%04d.html' % self.wrappers
        self.wrappers += 1
        self.add_bytes(filename,
                       parsers.IMAGE_WRAPPER.format(src=img_url,
                                                    title=img_title,
                                                    backlink="",
                                                    wrapper_class='x-ebookmaker-wrapper',
                                                    doctype=gg.XHTML_DOCTYPE),
                       mt.xhtml)
        return filename


class AdobePageMap(object):
    """ Class that builds a page-map xml file. """

    def __init__(self, ncx):
        self.toc = ncx.toc


    def __str__(self):
        return self.__unicode__()


    def __unicode__(self):
        """ Serialize page-map as unicode string. """

        pm = ElementMaker(namespace=str(NS.opf),
                          nsmap={None: str(NS.opf)})

        root = pm('page-map')

        for href, name, depth in self.toc:
            if depth == -1:
                root.append(pm.page(name=name, href=href))

        page_map = "%s\n\n%s" % (gg.XML_DECLARATION,
                                 etree.tostring(root,
                                                encoding=str,
                                                pretty_print=True))
        if options.verbose >= 3:
            debug(page_map)
        return page_map


class OutlineFixer(object):
    """ Class that fixes outline levels. """

    def __init__(self):
        self.stack = [(0, 1),]
        self.last = 1

    def level(self, in_level):
        if in_level < 1:
            return in_level
        (promotion, from_level) = self.stack[-1]
        if in_level > self.last + 1:
            # needs promotion
            more_promotion = in_level - self.last - 1
            new_promotion = promotion + more_promotion
            self.last = in_level
            self.stack.append((new_promotion, in_level))
            return in_level - new_promotion

        if in_level < from_level:
            # close out promotion
            self.last = from_level - promotion - 1
            self.stack.pop()
            return self.level(in_level)

        self.last = in_level
        return in_level - promotion


class TocNCX(object):
    """ Class that builds toc.ncx. """

    def __init__(self, dc):
        self.toc = []
        self.dc = dc
        self.seen_urls = {}
        self.ncx = ElementMaker(namespace=str(NS.ncx),
                                nsmap={None: str(NS.ncx)})


    def __str__(self):
        return self.__unicode__()


    def normalize_toc(self):
        """ normalize toc so that it starts with an h1 and doesn't jump down more than one """
        # level at a time
        fixer = OutlineFixer()
        top_level = 10
        for t in self.toc:
            t[2] = fixer.level(t[2])
            if t[2] > 0:
                top_level = min(t[2], top_level)
        if top_level != 1:
            for t in self.toc:
                t[2] = t[2] - top_level + 1

        # flatten toc if it contains only one top-level entry
        top_level_entries = list(filter(lambda x: x[2] == 1, self.toc))
        if len(top_level_entries) < 2:
            for t in self.toc:
                if t[2] != -1:
                    t[2] = max(1, t[2] - 1)

        # remove entries before the first level 1 entry
        if self.toc[0][2] != 1:
            self.toc = self.toc[self.toc.index(top_level_entries[0]):]

    def __unicode__(self):
        """ Serialize toc.ncx as unicode string. """
        ncx = self.ncx
        tocdepth = 1

        if self.toc:
            self.normalize_toc()

            tocdepth = max(t[2] for t in self.toc)

        head = ncx.head(
            ncx.meta(name='dtb:uid', content=self.dc.opf_identifier),
            ncx.meta(name='dtb:depth', content=str(tocdepth)),
            ncx.meta(name='dtb:generator', content=GENERATOR % VERSION),
            ncx.meta(name='dtb:totalPageCount', content='0'),
            ncx.meta(name='dtb:maxPageNumber', content='0'))

        doc_title = ncx.docTitle(ncx.text(self.dc.title))

        self.seen_urls = {}
        has_pages = False
        for url, dummy_title, depth in self.toc:
            # navPoints and pageTargets referencing the same element
            # must have the same playOrder
            if url not in self.seen_urls:
                self.seen_urls[url] = str(len(self.seen_urls) + 1)
            if depth == -1:
                has_pages = True

        params = {'version': '2005-1'}
        if self.dc.languages:
            params[NS.xml.lang] = self.dc.languages[0].id

        ncx = ncx.ncx(
            head,
            doc_title,
            self._make_navmap(self.toc),
            **params)

        if has_pages:
            ncx.append(self._make_pagelist(self.toc))

        # Ugly workaround for error: "Serialisation to unicode must not
        # request an XML declaration"

        toc_ncx = "%s\n\n%s" % (gg.XML_DECLARATION,
                                etree.tostring(ncx,
                                               doctype=gg.NCX_DOCTYPE,
                                               encoding=str,
                                               pretty_print=True))
        if options.verbose >= 3:
            debug(toc_ncx)
        return toc_ncx


    def rewrite_links(self, f):
        """ Rewrite all links f(). """
        for entry in self.toc:
            entry[0] = f(entry[0])


    def _make_navmap(self, toc):
        """ Build the toc. """
        ncx = self.ncx

        root = ncx.navMap()
        last_np_with_depth = {0: root}

        count = 0
        for url, title, depth in toc:
            if depth > -1:
                count += 1
                np = ncx.navPoint(
                    ncx.navLabel(ncx.text(title)),
                    ncx.content(src=url),
                    **{'id': "np-%d" % count,
                       'playOrder': self.seen_urls[url]})

                try:
                    parent = last_np_with_depth[depth - 1]
                    parent.append(np)
                    last_np_with_depth[depth] = np
                except KeyError:
                    warning("Bogus depth %d in TOC" % depth)

        return root


    def _make_pagelist(self, toc):
        """ Build the page list. """
        ncx = self.ncx

        root = ncx.pageList(
            ncx.navLabel(ncx.text('Pages')),
            # both attributes are optional in the spec,
            # but epubcheck 1.0.3 requires them
            **{'id': 'pages',
               'class': 'pagelist'})

        for url, pagename, depth in toc:
            if depth == -1:
                pt = ncx.pageTarget(
                    ncx.navLabel(ncx.text(pagename)),
                    ncx.content(src=url),
                    **{'id': "pt-%d" % len(root),
                       'value': str(len(root)), # fixme: extract value
                       'type': 'normal' if re.search('[0-9]', pagename) else 'front',
                       'playOrder': self.seen_urls[url]})

                root.append(pt)

        return root


class ContentOPF(object):
    """ Class that builds content.opf metadata. """

    def __init__(self):
        self.nsmap = gg.build_nsmap('opf dc dcterms xsi')

        # FIXME: remove this when lxml is fixed
        # workaround for lxml fat-fingering the default attribute namespaces
        self.nsmap[None] = str(NS.opf) + 'lxml-bug-workaround'

        self.opf = ElementMaker(namespace=self.nsmap[None], nsmap=self.nsmap)

        self.metadata = self.opf.metadata()
        self.manifest = self.opf.manifest()
        self.spine = self.opf.spine()
        self.guide = self.opf.guide()
        self.item_id = 0


    def __str__(self):
        return self.__unicode__()


    def __unicode__(self):
        """ Serialize content.opf as unicode string. """

        assert len(self.manifest) > 0, 'No manifest item in content.opf.'
        assert len(self.spine) > 0, 'No spine item in content.opf.'
        assert 'toc' in self.spine.attrib, 'No TOC item in content.opf.'

        package = self.opf.package(
            **{'version': '2.0', 'unique-identifier': 'id'})
        package.append(self.metadata)
        package.append(self.manifest)
        package.append(self.spine)
        if len(self.guide) > 0:
            package.append(self.guide)

        content_opf = "%s\n\n%s" % (gg.XML_DECLARATION,
                                    etree.tostring(package,
                                                   encoding=str,
                                                   pretty_print=True))

        # FIXME: remove this when lxml is fixed
        # now merge xmlns:opf and xmlns:
        content_opf = content_opf.replace('lxml-bug-workaround', '')

        if options.verbose >= 3:
            debug(content_opf)
        return content_opf


    def rewrite_links(self, f):
        """ Rewrite all links through f(). """
        for item in self.manifest:
            if item.get('href'):
                item.set('href', f(item.get('href')))
        for item in self.guide:
            item.set('href', f(item.get('href')))


    def guide_item(self, url, type_, title):
        """ Add item to guide. """
        self.guide.append(
            self.opf.reference(type=type_, title=title, href=url))


    def meta_item(self, name, content):
        """ Add item to metadata. """
        self.metadata.append(self.opf.meta(name=name, content=content))


    def manifest_item(self, url, mediatype, id_=None):
        """ Add item to manifest. """

        if id_ is None or xpath(self.manifest, "//*[@id = '%s']" % id_):
            self.item_id += 1
            id_ = 'item%d' % self.item_id

        self.manifest.append(
            self.opf.item(**{
                'href': url,
                'id': id_,
                'media-type': mediatype}))

        return id_


    def spine_item(self, url, mediatype, id_=None, linear=True, first=False):
        """ Add item to spine and manifest. """
        linear = 'yes' if linear else 'no'

        if id_ and id_.startswith('pgepubid'):
            # this is an auto-generated header id, not human-readable and probably duplicated
            # make a new one
            id_ = None

        id_ = self.manifest_item(url, mediatype, id_)

        # HACK: ADE needs cover flow as first element
        # but we don't know if we have a native coverpage until the manifest is complete
        if first:
            self.spine.insert(
                0, self.opf.itemref(idref=id_, linear=linear))
        else:
            self.spine.append(
                self.opf.itemref(idref=id_, linear=linear))


    def manifest_item_from_parser(self, p):
        """ Add item to manifest from parser. """
        if hasattr(p.attribs, 'comment'):
            self.manifest.append(etree.Comment(p.attribs.comment))
        return self.manifest_item(p.attribs.url, p.mediatype(), p.attribs.id)


    def spine_item_from_parser(self, p):
        """ Add item to spine and manifest from parser. """
        if hasattr(p.attribs, 'comment'):
            self.manifest.append(etree.Comment(p.attribs.comment))
        return self.spine_item(p.attribs.url, p.mediatype(), p.attribs.id,
                               linear=not hasattr(p.attribs, 'nonlinear'))


    def toc_item(self, url):
        """ Add TOC to manifest and spine. """
        self.manifest_item(url, 'application/x-dtbncx+xml', 'ncx')
        self.spine.attrib['toc'] = 'ncx'


    def pagemap_item(self, url):
        """ Add page-map to manifest and spine. """
        self.manifest_item(url, 'application/oebps-page-map+xml', 'map')
        self.spine.attrib['page-map'] = 'map'


    def metadata_item(self, dc):
        """ Build metadata from DublinCore struct.

        Example of metadata:

  <metadata xmlns:dcterms='http://purl.org/dc/terms/'
            xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance'
            xmlns:opf='http://www.idpf.org/2007/opf'>

    <dcterms:identifier opf:scheme='URI' id='id'>http://www.gutenberg.org/ebooks/29000</dc:identifier>
    <dcterms:creator opf:file-as='Trollope, Anthony'>Anthony Trollope</dc:creator>
    <dcterms:title>The Macdermots of Ballycloran</dc:title>
    <dcterms:language xsi:type='dcterms:RFC3066'>en</dc:language>
    <dcterms:subject>Domestic fiction</dc:subject>
    <dcterms:subject>Ireland -- Fiction</dc:subject>
    <dcterms:created>1890</dcterms:created>
    <dcterms:publisher>Project Gutenberg</dc:publisher>
    <dcterms:date opf:event='publication'>2009-05-31</dc:date>
    <dcterms:date opf:event='conversion'>2009-08-26T21:11:14Z</dc:date>
    <dcterms:rights>Public domain</dc:rights>
    <dcterms:source>29000-h.htm</dc:source>

    <meta name='cover' content='item0' />
  </metadata>
    """

        # OPF 2.0 v1.0 specifies to use the
        # Dublin Core Metadata Element Set, Version 1.1
        # http://dublincore.org/documents/2004/12/20/dces/
        # but that has been superseded by DCMI Metadata Terms
        # http://dublincore.org/documents/dcmi-terms/
        # we use NS.dc for now but should switch to NS.dcterms later

        dcterms = ElementMaker(nsmap=self.nsmap, namespace=str(NS.dc))

        if dc.publisher:
            self.metadata.append(dcterms.publisher(dc.publisher))
        if dc.rights:
            self.metadata.append(dcterms.rights(dc.rights))

        self.metadata.append(dcterms.identifier(dc.opf_identifier, {
            NS.opf.scheme: 'URI',
            'id': 'id'})) # should be NS.xml.id

        for author in dc.authors:
            pretty_name = dc.make_pretty_name(author.name)
            if author.marcrel == 'aut' or author.marcrel == 'cre':
                self.metadata.append(dcterms.creator(
                    pretty_name, {NS.opf['file-as']: author.name}))
            else:
                self.metadata.append(dcterms.contributor(
                    pretty_name, {NS.opf.role: author.marcrel,
                                  NS.opf['file-as']: author.name}))

        # replace newlines with /
        title = re.sub(r'\s*[\r\n]+\s*', ' / ', dc.title)
        self.metadata.append(dcterms.title(title))

        for language in dc.languages:
            self.metadata.append(dcterms.language(
                language.id, {NS.xsi.type: 'dcterms:RFC4646'}))

        for subject in dc.subjects:
            self.metadata.append(dcterms.subject(subject.subject))

        if dc.created:
            self.metadata.append(dcterms.date(
                dc.created, {NS.opf.event: 'creation'}))

        if dc.release_date != datetime.date.min:
            self.metadata.append(dcterms.date(
                dc.release_date.isoformat(),
                {NS.opf.event: 'publication'}))

        self.metadata.append(dcterms.date(
            datetime.datetime.now(gg.UTC()).isoformat(),
            {NS.opf.event: 'conversion'}))

        source = dc.source
        if hasattr(options.config, 'FILESDIR'):
            if source.startswith(options.config.FILESDIR):
                source = source[len(options.config.FILESDIR):]
                source = urllib.parse.urljoin(options.config.PGURL, source)

        self.metadata.append(dcterms.source(source))


    def add_coverpage(self, ocf, url):
        """ Add a coverpage for ADE and Kindle.

        The recommended cover size is 600x800 pixels (500 pixels on
        the smaller side is an absolute minimum). The cover page
        should be a color picture in JPEG format.

        """

        id_ = None

        # look for a manifest item with the right url
        for item in xpath(
                self.manifest,
                # cannot xpath for default namespace
                "//*[local-name () = 'item' and (starts-with (@media-type, 'image/jpeg') or starts-with(@media-type, 'image/png')) and @href = $url]",
                url=url):

            id_ = item.get('id')
            break

        # else use default cover page image
        if id_ is None:
            ext = url.split('.')[-1]
            try:
                mediatype = getattr(mt, ext)
            except AttributeError:
                mediatype = mt.jpeg
            try:
                with open(url, 'r') as f:
                    ocf.add_bytes(Writer.url2filename(url), f.read(), mediatype)
            except IOError:
                url = 'cover.jpg'
                ocf.add_bytes(url, resource_string('ebookmaker.writers', url), mediatype)
            id_ = self.manifest_item(url, mediatype)

        debug("Adding coverpage id: %s url: %s" % (id_, url))

        # register mobipocket style
        self.meta_item('cover', id_)

        # register ADE style
        href = ocf.add_image_wrapper(Writer.url2filename(url), 'Cover')
        self.spine_item(href, mt.xhtml, 'coverpage-wrapper', True, True)
        self.guide_item(href, 'cover', 'Cover')



class Writer(writers.HTMLishWriter):
    """ Class that writes epub files. """

    VALIDATOR = 'EPUB_VALIDATOR'

    @staticmethod
    def strip_pagenumbers(xhtml, strip_classes):
        """ Strip dp page numbers.

        Rationale: DP implements page numbers either with float or
        with absolute positioning. Float is not supported by Kindle.
        Absolute positioning is not allowed in epub.

        If we'd leave these in, they would show up as numbers in the
        middle of the text.

        To still keep links working, we replace all page number
        contraptions we can find with empty <a>'s.

        """

        # look for elements with a class that is in strip_classes

        for class_ in strip_classes:
            xp = "//xhtml:*[@class and contains(concat(' ', normalize-space(@class), ' '), ' %s ')]" % class_

            count = 0
            for elem in xpath(xhtml, xp):

                # save textual content
                text = gg.normalize(etree.tostring(elem,
                                                   method="text",
                                                   encoding=str,
                                                   with_tail=False))
                if len(text) > 10:
                    # safeguard against removing things that are not pagenumbers
                    continue

                if not text:
                    text = elem.get('title')

                # look for id anywhere inside element
                id_ = elem.xpath(".//@id")

                # transmogrify element into empty <a> or <div>
                tail = elem.tail
                elem.clear()
                if elem.getparent().tag == NS.xhtml.body:
                    elem.tag = NS.xhtml.div
                else:
                    elem.tag = NS.xhtml.a
                if id_:
                    # some blockheaded PPers include more than
                    # one page number in one span. take the last id
                    # because the others represent empty pages.
                    elem.set('id', id_[-1])

                if class_ in DP_PAGENUMBER_CLASSES:
                    # mark element as rewritten pagenumber. we
                    # actually don't use this class for styling
                    # because it is on an empty element
                    elem.set('class', 'x-ebookmaker-pageno')

                if text:
                    elem.set('title', text)
                elem.tail = tail
                count += 1

                # The OPS Spec 2.0 is very clear: "Reading Systems
                # must be XML processors as defined in XML 1.1."
                # Nevertheless many browser-plugin ebook readers use
                # the HTML parsers of the browser.  But HTML parsers
                # don't grok the minimized form of empty elements.
                #
                # This will force lxml to output the non-minimized form
                # of the element.
                elem.text = ''

            if count:
                warning("%d elements having class %s have been rewritten." %
                        (count, class_))


    @staticmethod
    def insert_root_div(xhtml):
        """ Insert a div immediately below body and move body contents
        into it.

        Rationale: We routinely turn page numbers into <a> elements.
        <a> elements are illegal as children of body, but are legal as
        children of <div>. See: `strip_page_numbers ()`

        """
        em = ElementMaker(namespace=str(NS.xhtml),
                          nsmap={None: str(NS.xhtml)})

        for body in xpath(xhtml, "/xhtml:body"):
            div = em.div
            div.set('id', 'pgepub-root-div')
            for child in body:
                div.append(child)
            body.append(div)


    # characters that are not widely supported
    translate_map = {
        0x2012: 0x2013,    # U+2012 FIGURE-DASH    -> U+2013 EN-DASH (ADE lacks this)
        0x2015: 0x2014,    # U+2015 HORIZONTAL BAR -> U+2014 EM-DASH (ADE lacks this)
    }

    @staticmethod
    def fix_charset(xhtml):
        """ Replace some characters that are not widely supported. """

        for node in xhtml.iter():
            if node.text:
                node.text = str(node.text).translate(Writer.translate_map)
            if node.tail:
                node.tail = str(node.tail).translate(Writer.translate_map)


    @staticmethod
    def fix_incompatible_css(sheet):
        """ Strip CSS properties and values that are not EPUB compatible.
            Unpack "media handheld" rules
        """

        cssclass = re.compile(r'\.(-?[_a-zA-Z]+[_a-zA-Z0-9-]*)')
        html5tag = re.compile(r'(^|[ ,~>+])(figure|figcaption|footer|header|section)')

        for rule in sheet:
            if rule.type == rule.MEDIA_RULE:
                if rule.media.mediaText.find('handheld') > -1:
                    debug("Unpacking CSS @media handheld rule.")
                    rule.media.mediaText = 'all'
                    info("replacing  @media handheld rule with @media all")

            if rule.type == rule.STYLE_RULE:
                #change html5 tags to classes with the same name
                newrule = html5tag.sub(r'\1div.\2', rule.selectorList.selectorText)
                rule.selectorList.selectorText = newrule

                ruleclasses = list(cssclass.findall(rule.selectorList.selectorText))
                for p in list(rule.style):
                    if p.name == 'float' and "x-ebookmaker" not in ruleclasses:
                        debug("Dropping property %s" % p.name)
                        rule.style.removeProperty('float')
                        rule.style.removeProperty('width')
                        rule.style.removeProperty('height')
                    elif p.name == 'position':
                        debug("Dropping property %s" % p.name)
                        rule.style.removeProperty('position')
                        rule.style.removeProperty('left')
                        rule.style.removeProperty('right')
                        rule.style.removeProperty('top')
                        rule.style.removeProperty('bottom')
                    elif p.name in ('background-image', 'background-position',
                                    'background-attachment', 'background-repeat'):
                        debug("Dropping property %s" % p.name)
                        rule.style.removeProperty(p.name)
                    elif 'border' not in p.name and 'px' in p.value:
                        debug("Dropping property with px value %s" % p.name)
                        rule.style.removeProperty(p.name)
                    elif p.name == 'speak':
                        val = rule.style.getPropertyValue('speak')
                        if val == 'spell-out':
                            rule.style.removeProperty('speak')
                            rule.style.setProperty('speak-as', 'spell-out')
                        elif val == 'none':
                            rule.style.setProperty('speak', 'never')
                        elif val == 'inherit':
                            rule.style.setProperty('speak', 'auto')

        # debug("exit fix_incompatible_css")


    @staticmethod
    def get_classes_with_prop(xhtml, props=('float', 'position')):
        """ Get a list of all classes that use float or position. """

        classes = set()
        regex = re.compile(r"\.(\w+)", re.ASCII)

        for style in xpath(xhtml, "//xhtml:style"):
            p = parsers.CSSParser.Parser()
            if style.text: # try to fix os-dependent empty style bug
                p.parse_string(style.text)

                for rule in p.sheet:
                    if rule.type == rule.STYLE_RULE:
                        for p in rule.style:
                            if p.name in props:
                                classes.update(regex.findall(rule.selectorList.selectorText))
                                break

        return classes


    @staticmethod
    def fix_style_elements(xhtml):
        """ Fix CSS style elements.  Make sure they are utf-8. """

        # debug("enter fix_style_elements")

        for style in xpath(xhtml, "//xhtml:style"):
            p = parsers.CSSParser.Parser()
            if style.text: # try to fix os-dependent empty style bug
                p.parse_string(style.text)
                try:
                    # pylint: disable=E1103
                    style.text = p.sheet.cssText.decode('utf-8')
                except (ValueError, UnicodeError):
                    debug("CSS:\n%s" % p.sheet.cssText)
                    raise

        # debug("exit fix_style_elements")

    @staticmethod
    def render_q(xhtml):
        """ replace q elements with span surrounded by curly quotes. """
        def surround(elem, before, after):
            elem.text = before + elem.text if elem.text else before
            if len(elem) > 0:
                elem[-1].tail = elem[-1].tail + after if elem[-1].tail else after
            else:
                elem.text = elem.text +after

        # nested quotes
        for q in xpath(xhtml, "//xhtml:q//xhtml:q"):
            q.tag = "span"
            surround(q, '‘', '’')
        # un-nested quotes
        for q in xpath(xhtml, "//xhtml:q"):
            q.tag = "span"
            surround(q, '“', '”')

    @staticmethod
    def fix_html5(xhtml):
        """
        Find html5 constructs and tries to fix them for xhtml
        """
        #remove some tags
        for meta in xpath(xhtml, '//xhtml:meta[@charset]'):
            meta.getparent().remove(meta)

        for newtag in ['wbr',]:
            for tag in xpath(xhtml, f'//xhtml:{newtag}'):
                tag.getparent().remove(tag)

        for meta in xpath(xhtml, '//xhtml:meta[@property]'):
            meta.getparent().remove(meta)

        # html5 moved tfoot to end of the table
        for tfoot in xpath(xhtml, '//xhtml:*[last()][name()="tfoot"]'):
            tbody = tfoot.getparent().find('{*}tbody')
            if tbody is not None:
                tbody.addprevious(tfoot)

        # set required attributes removed in html5
        attrs_to_fill = [('style', 'type', 'text/css'), ('table', 'summary', '')]
        for (tag, attr, fill) in attrs_to_fill:
            for elem in xpath(xhtml, f"//xhtml:{tag}[not(@{attr})]"):
                elem.set(attr, fill)

        # remove html5-only attribute
        attrs_to_remove = [('*', 'role'),('*', 'aria-label'),('*', 'aria-labelledby'),]
        for (tag, attr) in attrs_to_remove:
            for elem in xpath(xhtml, f"//xhtml:{tag}[@{attr}]"):
                del elem.attrib[attr]

        usedtags = set()
        for newtag in ['figcaption', 'figure', 'footer', 'header', 'section']:
            for tag in xpath(xhtml, f'//xhtml:{newtag}'):
                usedtags.add(newtag)
                tag.tag = NS.xhtml.div
                writers.HTMLWriter.add_class(tag, newtag)

        if 'figure' in usedtags:
            Writer.add_internal_css(xhtml, 'div.figure {margin: 1em 40px;}')     


    @staticmethod
    def strip_links(xhtml, manifest):
        """
        Strip all links to local resources that aren't in manifest or are images.

        This does not strip inline images, only standalone images that
        are targets of links. EPUB does not allow that.

        """

        for link in xpath(xhtml, '//xhtml:a[@href]'):
            href = urllib.parse.urldefrag(link.get('href'))[0]
            if href in manifest and not manifest[href].startswith('image'):
                continue
            if not href.startswith('file:'):
                continue
            debug("strip_links: Deleting <a> to file not in manifest: %s" % href)
            del link.attrib['href']


    @staticmethod
    def strip_ins(xhtml):
        """
        Strip all <ins> tags.

        There's a bug in the epub validator that trips on class and
        title attributes in <ins> elements.

        """
        for ins in xpath(xhtml, '//xhtml:ins'):
            ins.drop_tag()


    @staticmethod
    def strip_noepub(xhtml):
        """ Strip all <* class='x-ebookmaker-drop'> tags.

        As a way to tailor your html towards epub.

        """

        for e in xpath(xhtml, "//xhtml:*[contains (@class, 'x-ebookmaker-drop')]"):
            # preserve any ids ing the tree we're dropping
            dropped_ids = []
            if 'id' in e.attrib:
                dropped_ids.append(e.attrib['id'])
            for id_el in xpath(e, './/xhtml:*[@id]'):
                dropped_ids.append(id_el.attrib['id'])
            for dropped_id in dropped_ids:
                newtag = NS.xhtml.div if e.getparent().tag == NS.xhtml.body else NS.xhtml.span
                e.addprevious(xhtml.makeelement(newtag, attrib={'id': dropped_id}))
            e.drop_tree()


    @staticmethod
    def strip_rst_dropcaps(xhtml):
        """ Replace <img class='dropcap'> with <span class='dropcap'>.

        """

        for e in xpath(xhtml, "//xhtml:img[@class ='dropcap']"):
            e.tag = NS.xhtml.span
            e.text = e.get('alt', '')

    @staticmethod
    def strip_data_attribs(xhtml):
        """ Epubcheck doesn't like data- attributes in EPUB2.
        """
        for e in xpath(xhtml, "//@*[starts-with(name(), 'data')]/.."):
            for key in e.attrib.keys():
                if key.startswith('data-'):
                    del e.attrib[key]


    @staticmethod
    def reflow_pre(xhtml):
        """ make <pre> reflowable.

        This helps a lot with readers like Sony's that cannot
        scroll horizontally.

        """

        def nbsp(matchobj):
            return (' ' * (len(matchobj.group(0)) - 1)) + ' '

        for pre in xpath(xhtml, "//xhtml:pre"):
            # white-space: pre-wrap would do fine
            # but it is not supported by OEB
            try:
                pre.tag = NS.xhtml.div
                writers.HTMLishWriter.add_class(pre, 'pgmonospaced')
                try:
                    m = parsers.RE_GUTENBERG.search(pre.text)
                    if m:
                        writers.HTMLishWriter.add_class(pre, 'pgheader')
                except TypeError:
                    pass
                tail = pre.tail
                s = etree.tostring(pre, encoding=str, with_tail=False)
                s = s.replace('>\n', '>')      # eliminate that empty first line
                s = s.replace('\n', '\n<br/>')
                s = re.sub('  +', nbsp, s)
                div = etree.fromstring(s)
                div.tail = tail

                pre.getparent().replace(pre, div)

            except etree.XMLSyntaxError as what:
                exception("%s\n%s" % (s, what))
                raise


    @staticmethod
    def single_child(e):
        """ Resturn true if node contains a single child element and nothing else. """
        p = e.getparent()
        return (len(p) == 1 and
                (p.text is None or p.text.isspace()) and
                (e.tail is None or e.tail.isspace()))


    @staticmethod
    def url2filename(url):
        """ Generate a filename for this url.
            - preserve original filename and fragment
            - map directory path to a cross platform filename string

        """
        if match_link_url.match(url):
            return url
        if url.startswith('file://'):
            url = url[7:]

        url_match = match_non_link.search(url)
        prefix = url[0:-len(url_match.group(0))]
        if prefix:
            prefix = abs(hash(prefix))
            return f'{prefix}_{url_match.group(0)}'
        return url


    @staticmethod
    def rescale_into(dimen, max_dimen):
        """ Scale down dimen to fit into max_dimen. """
        scale = 1.0
        if dimen[0]:
            scale = min(scale, max_dimen[0] / float(dimen[0]))
        if dimen[1]:
            scale = min(scale, max_dimen[1] / float(dimen[1]))

        if scale < 1.0:
            dimen = (int(dimen[0] * scale) if dimen[0] else None,
                     int(dimen[1] * scale) if dimen[1] else None)

        return dimen


    @staticmethod
    def fix_html_image_dimensions(xhtml):
        """
        Remove all width and height that is not specified in '%'.
        """

        for img in xpath(xhtml, '//xhtml:img'):
            a = img.attrib

            if '%' in a.get('width', '%') and '%' in a.get('height', '%'):
                continue

            if 'width' in a:
                del a['width']
            if 'height' in a:
                del a['height']


    def remove_coverpage(self, xhtml, url):
        """ Remove coverpage from flow.

        EPUB readers will display the coverpage from the manifest and
        if we don't remove it from flow it will be displayed twice.

        """
        for img in xpath(xhtml, "//xhtml:img[@src = $url and not(contains(@class, 'x-ebookmaker-important'))]", url=url):
            debug("remove_coverpage: dropping <img> %s from flow" % url)
            img.drop_tree()
            return # only the first one though


    def shipout(self, job, parserlist, ncx):
        """ Build the zip file. """

        try:
            ocf = OEBPSContainer(
                os.path.join(os.path.abspath(job.outputdir), job.outputfile),
                ('%d/' % options.ebook if options.ebook else None))

            opf = ContentOPF()

            opf.metadata_item(job.dc)

            # write out parserlist

            for p in parserlist:
                try:
                    ocf.add_bytes(self.url2filename(p.attribs.url), p.serialize(),
                                  p.mediatype())
                    if p.mediatype() == mt.xhtml:
                        opf.spine_item_from_parser(p)
                    else:
                        opf.manifest_item_from_parser(p)
                except Exception as what:
                    error("Could not process file %s: %s" % (p.attribs.url, what))

            # toc

            for t in ncx.toc:
                if t[1].lower().strip(' .') in TOC_HEADERS:
                    opf.guide_item(t[0], 'toc', t[1])
                    break

            opf.toc_item('toc.ncx')
            ocf.add_unicode('toc.ncx', str(ncx))

            for p in parserlist:
                if 'icon' in p.attribs.rel:
                    opf.add_coverpage(ocf, p.attribs.url)
                    break

            opf.rewrite_links(self.url2filename)
            ocf.add_unicode('content.opf', str(opf))

            ocf.commit()

        except Exception as what:
            exception("Error building Epub: %s" % what)
            ocf.rollback()
            raise


    def build(self, job):
        """ Build epub """

        ncx = TocNCX(job.dc)
        parserlist = []
        css_count = 0

        # add CSS parser
        self.add_external_css(job.spider, None, PRIVATE_CSS, 'pgepub.css')

        try:
            chunker = HTMLChunker.HTMLChunker()
            coverpage_url = None

            # do images early as we need the new dimensions later
            for p in job.spider.parsers:
                if hasattr(p, 'resize_image'):
                    if 'icon' in p.attribs.rel:
                        if job.subtype == '.noimages':
                            np = p.resize_image(MAX_NOIMAGE_SIZE, MAX_COVER_DIMEN)
                        else:
                            np = p.resize_image(MAX_IMAGE_SIZE, MAX_COVER_DIMEN)
                        np.id = p.attribs.get('id', 'coverpage')
                        coverpage_url = p.attribs.url
                    elif 'linked_image' in p.attribs.rel:
                        if job.subtype == '.noimages':
                            np = p.resize_image(MAX_NOIMAGE_SIZE, LINKED_IMAGE_DIMEN)
                        else:
                            np = p.resize_image(LINKED_IMAGE_SIZE, LINKED_IMAGE_DIMEN)
                        np.id = p.attribs.get('id')
                    else:
                        np = p.resize_image(MAX_IMAGE_SIZE, MAX_IMAGE_DIMEN)
                        np.id = p.attribs.get('id')
                    parserlist.append(np)

            for p in job.spider.parsers:
                if p.mediatype() in OPS_CONTENT_DOCUMENTS:
                    debug("URL: %s" % p.attribs.url)

                    if hasattr(p, 'rst2epub2'):
                        xhtml = p.rst2epub2(job)
                        xhtml = copy.deepcopy(xhtml)

                        if options.verbose >= 2:
                            # write html to disk for debugging
                            debugfilename = os.path.join(os.path.abspath(job.outputdir),
                                                         job.outputfile)
                            debugfilename = os.path.splitext(debugfilename)[0] + '.' + \
                                job.maintype + '.debug.html'
                            with open(debugfilename, 'wb') as fp:
                                fp.write(etree.tostring(xhtml, encoding='utf-8'))

                    else:
                        # make a copy so we can mess around
                        p.parse()
                        xhtml = copy.deepcopy(p.xhtml) if hasattr(p, 'xhtml') else None
                    if xhtml is not None:
                        self.fix_html5(xhtml)

                        strip_classes = self.get_classes_with_prop(xhtml)
                        strip_classes = strip_classes.intersection(STRIP_CLASSES)
                        if strip_classes:
                            self.strip_pagenumbers(xhtml, strip_classes)

                        # build up TOC
                        # has side effects on xhtml
                        ncx.toc += p.make_toc(xhtml)

                        # allows authors to customize css for epub
                        self.add_body_class(xhtml, 'x-ebookmaker')
                        self.add_body_class(xhtml, 'x-ebookmaker-2')

                        self.insert_root_div(xhtml)
                        self.fix_charset(xhtml)
                        self.fix_style_elements(xhtml)

                        if job.subtype in ('.images', '.noimages'):
                            # omit for future subtype '.v3'
                            self.reflow_pre(xhtml)
                            self.render_q(xhtml)


                        # strip all links to items not in manifest
                        p.strip_links(xhtml, job.spider.dict_urls_mediatypes())
                        self.strip_links(xhtml, job.spider.dict_urls_mediatypes())
                        self.strip_data_attribs(xhtml)

                        self.strip_noepub(xhtml)
                        # self.strip_rst_dropcaps(xhtml)

                        self.fix_html_image_dimensions(xhtml)
                        if coverpage_url and not hasattr(p.attribs, 'nonlinear'):
                            self.remove_coverpage(xhtml, coverpage_url)

                        # externalize and fix CSS
                        for style in xpath(xhtml, '//xhtml:style'):
                            self.add_external_css(
                                job.spider, xhtml, style.text, "%d.css" % css_count)
                            css_count += 1
                            style.drop_tree()

                        self.add_external_css(job.spider, xhtml, None, 'pgepub.css')

                        self.add_meta_generator(xhtml)

                        debug("Splitting %s ..." % p.attribs.url)
                        chunker.next_id = 0
                        chunker.split(xhtml, p.attribs)
                    else:
                        # parsing xml worked, but it isn't xhtml. so we need to reset mediatype
                        # to something that isn't recognized as content
                        p.attribs.mediatype = parsers.ParserAttributes.HeaderElement('text/xml')
            for p in job.spider.parsers:
                if str(p.attribs.mediatype) == 'text/css':
                    p.parse()
                if hasattr(p, 'sheet') and p.sheet:
                    self.fix_incompatible_css(p.sheet)
                    if job.subtype == '.noimages':
                        p.strip_images()
                    p.rewrite_links(self.url2filename)
                    parserlist.append(p)
                if str(p.attribs.mediatype) in OPS_FONT_TYPES:
                    warning('font file embedded: %s ;  check its license!', p.attribs.url)
                    parserlist.append(p)

            # after splitting html into chunks we have to rewrite all
            # internal links in HTML
            chunker.rewrite_internal_links()
            # also in the TOC
            if not ncx.toc:
                ncx.toc.append([job.spider.parsers[0].attribs.url, 'Start', 1])
            chunker.rewrite_internal_links_toc(ncx.toc)

            # make absolute links zip-filename-compatible
            chunker.rewrite_links(self.url2filename)
            ncx.rewrite_links(self.url2filename)

            # Do away with the chunker and copy all chunks into new parsers.
            # These are fake parsers that never actually parsed anything,
            # we just use them to just hold our data.
            for chunk, attribs in chunker.chunks:
                p = ParserFactory.ParserFactory.get(attribs)
                p.xhtml = chunk
                parserlist.append(p)

            self.shipout(job, parserlist, ncx)

        except Exception as what:
            exception("Error building Epub: %s" % what)
            raise



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""

parts.py

Copyright 2010-2012 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

New and tweaked transformers.

"""

from __future__ import unicode_literals

import re
import fnmatch

import six

from docutils import nodes
import docutils.transforms
import docutils.transforms.parts

from ebookmaker.mydocutils import nodes as mynodes
from libgutenberg.Logger import error, warning, info, debug
from ebookmaker import Unitame

# pylint: disable=W0142

def copy_and_filter (node, document):
    """ Return a copy of a title, with references, images, etc. removed. """

    visitor = ContentsFilter (document)
    node.walkabout (visitor)
    return visitor.get_entry_text ()


class ContentsFilter (nodes.TreeCopyVisitor):

    def get_entry_text (self):
        return self.get_tree_copy ().children

    def ignore_node (self, node):
        raise nodes.SkipNode

    def ignore_node_but_process_children (self, node):
        raise nodes.SkipDeparture

    def visit_image (self, node):
        if node.hasattr ('alt'):
            self.parent.append (nodes.Text (node['alt']))
        raise nodes.SkipNode

    visit_newline = nodes.TreeCopyVisitor.default_visit
    depart_newline = nodes.TreeCopyVisitor.default_departure

    visit_citation_reference = ignore_node
    visit_footnote_reference = ignore_node
    visit_raw = ignore_node

    visit_interpreted = ignore_node_but_process_children
    visit_problematic = ignore_node_but_process_children
    visit_reference = ignore_node_but_process_children
    visit_target = ignore_node_but_process_children


###########
### 200 ###
###########


class Lineblock2VSpace (docutils.transforms.Transform):
    """
    Turn empty line_blocks into page nodes.

    """

    default_priority = 200 # early, before vspace

    def apply (self):
        for lb in self.document.traverse (nodes.line_block):
            if lb.astext ().strip () == '':
                gap_len = len (lb)
                page = mynodes.page ()
                page['classes'].append ('vspace')
                page['length'] = gap_len
                lb.replace_self (page)


###########
### 300 ###
###########


class DocInfoCollector (docutils.transforms.Transform):
    """
    Collects docinfo in document.meta_block (dict of lists).

    """

    default_priority = 341
    """ After frontmatter.py DocInfo """

    def apply (self):
        doc = self.document

        for docinfo in doc.traverse (nodes.docinfo):
            for field in docinfo:
                if isinstance (field, nodes.field):
                    field_name, field_body = field.children
                    doc.meta_block[field_name.astext ()].append (field_body.astext ())
                elif isinstance (field, nodes.Bibliographic):
                    doc.meta_block[field.__class__.__name__].append (field.astext ())

            docinfo.parent.remove (docinfo)


class TitleLevelTransform (docutils.transforms.Transform):
    """
    Adds some useful classes to sections, titles and captions.

    Add `level-N` to `section`, `title` and `subtitle` nodes.

    Add `document-title` or `section-title`, or ...

    Add `title` or `subtitle`.

    Add `with-subtitle`.

    Add `figure-caption`.

    """

    default_priority = 361 # after title promotion

    def apply (self, **kwargs):
        self.recurse (self.document, 1)

        for caption in self.document.traverse (nodes.caption):
            caption['classes'].append (caption.parent.__class__.__name__ + '-caption')

    def recurse (self, parent, level):
        for node in parent:
            if isinstance (node, nodes.Text):
                continue

            if isinstance (node, (nodes.title, nodes.subtitle)):
                nclass  = node.__class__.__name__
                pclass  = parent.__class__.__name__
                classes = [nclass,
                           '%s-%s' % (pclass, nclass),
                           'level-%d' % level]
                if (isinstance (node, (nodes.title)) and
                    len (parent) >= 2 and
                    isinstance (parent[1], nodes.subtitle)):
                    classes.append ('with-subtitle')
                node['classes'].extend (classes)
                node['level'] = level

            if isinstance (node, nodes.section):
                node['classes'].append ('level-%d' % (level + 1))
                node['level'] = level + 1

                self.recurse (node, level + 1)
            else:
                self.recurse (node, level)


###########
### 700 ###
###########


class TocPageNumberTransform (docutils.transforms.Transform):
    """ Finds the page number all sections, tables and figures are in. """

    default_priority = 715 # before contents transform

    def apply (self, **kwargs):
        pageno = '' # pageno is actually a string

        for n in self.document.traverse ():
            if isinstance (n, nodes.target) and 'pageno' in n:
                pageno = n['pageno']
                # debug ("pageno: >%s<" % self.pageno)
            elif isinstance (n, (nodes.section, nodes.figure, nodes.table)):
                n['pageno'] = pageno


class TocEntryTransform (docutils.transforms.Transform):
    """ Moves data of pending node onto next header. """

    default_priority = 717 # before Contents transform

    def apply (self, **kwargs):
        # debug ("TocEntryTransform %s" % repr (self.startnode.details))

        iter_ = self.startnode.traverse (
            lambda x: isinstance (x, (nodes.title, nodes.caption)), ascend = 1, descend = 1)

        details = self.startnode.details
        for node in iter_:
            title = node
            title['toc_entry'] = details['content']

            # copy depth
            if 'toc_depth' in details:
                section = title.parent
                if isinstance (section, nodes.section):
                    section['toc_depth'] = details['toc_depth']
                    # debug ("Setting toc_depth: %d" % section['toc_depth'])
            break
        self.startnode.parent.remove (self.startnode)


class ContentsTransform (docutils.transforms.Transform):
    """ A modified contents transform that obeys contents-depth directives. """

    default_priority = 720

    def __init__ (self, document, startnode = None):
        docutils.transforms.Transform.__init__ (self, document, startnode)
        self.depth = startnode.details.get ('depth', six.MAXSIZE) if startnode else six.MAXSIZE
        self.toc_depth = six.MAXSIZE
        self.use_pagenos = 'page-numbers' in startnode.details
        self.maxlen = 0 # length of longest page no.

    def apply(self):
        details = self.startnode.details
        if 'local' in details:
            startnode = self.startnode.parent.parent
            while not (isinstance (startnode, nodes.section)
                       or isinstance (startnode, nodes.document)):
                # find the ToC root: a direct ancestor of startnode
                startnode = startnode.parent
        else:
            startnode = self.document

        contents = self.build_contents (startnode)
        if len (contents):
            self.startnode.replace_self (contents)
        else:
            self.startnode.parent.parent.remove (self.startnode.parent)

    def build_contents (self, node, level=0):
        # debug ('build_contents level %d' % level)

        details = self.startnode.details
        backlinks = details.get ('backlinks', self.document.settings.toc_backlinks)
        try:
            toc_id = self.startnode.parent['ids'][0]
        except:
            toc_id = None

        entries = []
        for n in node:
            if isinstance (n, nodes.section):
                if 'toc_depth' in n:
                    self.toc_depth = n['toc_depth']
                    # debug ("New toc_depth: %d" % self.toc_depth)
                if level < self.depth and level < self.toc_depth:
                    subsects = self.build_contents(n, level + 1)
                    title = n[0]
                    if not isinstance (title, nodes.title):
                        continue
                    # debug ('title: %s level: %d depth: %d' % (title, level, self.toc_depth))

                    pagenos = []
                    if self.use_pagenos and 'pageno' in n:
                        self.maxlen = max (self.maxlen, len (n['pageno']))
                        inline = nodes.inline ('', ' ' + n['pageno'])
                        inline['classes'].append ('toc-pageref')
                        pagenos.append (inline)

                    if 'toc_entry' in title:
                        container = title['toc_entry']
                        if container is None: # suppress toc entry if emtpy
                            continue
                        # debug ("Setting TOC entry")
                        entrytext = copy_and_filter (container, self.document)
                    else:
                        entrytext = copy_and_filter (title, self.document)

                    reference = nodes.reference (
                        '', '', refid = n['ids'][0], *entrytext)
                    ref_id = self.document.set_id (reference)

                    entry = nodes.paragraph ('', '', reference, *pagenos)
                    item = nodes.list_item ('', entry)
                    item['refid'] = n['ids'][0]
                    item['classes'].append ('toc-entry')
                    if 'level' in title:
                        item['level'] = title['level']
                        item['classes'].append ('level-%d' % (title['level']))
                    if (backlinks in ('entry', 'top')
                         and title.next_node (nodes.Referential) is None):
                        if backlinks == 'entry':
                            title['refid'] = ref_id
                        elif backlinks == 'top' and toc_id is not None:
                            title['refid'] = toc_id
                    item += subsects
                    entries.append (item)
        if entries:
            return nodes.bullet_list ('', *entries, **{'classes': ['compact', 'toc-list'],
                                                       'enumtype': 'none',
                                                       'pageno_maxlen': self.maxlen})
        else:
            return []


class ListOfAnythingTransform (docutils.transforms.Transform):
    """ Create a List Of Figures or List Of Tables etc. """

    default_priority = 718 # before empty section remover

    def __init__ (self, document, startnode = None):
        docutils.transforms.Transform.__init__ (self, document, startnode)
        self.maxlen = 0 # length of longest page no.

    def apply (self, **kwargs):
        entries = []
        details = self.startnode.details

        self.backlinks = details.get ('backlinks',
                                      self.document.settings.toc_backlinks)
        self.use_pagenos = 'page-numbers' in details
        directive_name = details.get ('directive_name', '')

        condition = None
        if 'selector' in details:
            condition = mynodes.node_selector (details['selector'])
        elif directive_name == 'lof':
            condition = nodes.figure
        elif directive_name == 'lot':
            condition = nodes.table

        if not condition:
            self.document.reporter.warning (
                "directive %s needs an option 'selector'" % directive_name,
                base_node=self.startnode.parent)
            raise docutils.transform.TransformError

        if 'local' in details:
            startnode = self.startnode.parent.parent
            while not isinstance (startnode, (nodes.section, nodes.document)):
                # find the ToC root: a direct ancestor of startnode
                startnode = startnode.parent
        else:
            startnode = self.document

        try:
            toc_id = self.startnode.parent.parent['ids'][0]
        except:
            toc_id = None

        for node in startnode.traverse (condition):

            title = list (node.traverse (nodes.caption)) + list (node.traverse (nodes.title))
            if len (title) != 1:
                # cannot put anonymous X in list of X
                continue

            title = title[0]

            pagenos = []
            if self.use_pagenos and 'pageno' in node:
                self.maxlen = max (self.maxlen, len (node['pageno']))
                inline = nodes.inline ('', ' ' + node['pageno'])
                inline['classes'].append ('toc-pageref')
                pagenos.append (inline)

            if 'toc_entry' in title:
                container = title['toc_entry']
                if container is None: # suppress toc entry if emtpy
                    # debug ("Suppressing TOC entry")
                    continue
                # debug ("Setting TOC entry")
                entrytext = copy_and_filter (container, self.document)
            else:
                entrytext = copy_and_filter (title, self.document) # returns list of nodes

            reference = nodes.reference ('', '', *entrytext, refid = node['ids'][0])
            ref_id = self.document.set_id (reference) # target for backlink

            list_item = nodes.list_item ('', nodes.inline ('', '', reference, *pagenos))
            list_item['refid'] = node['ids'][0]
            list_item['level'] = 2
            list_item['classes'].append ('level-2')
            list_item['classes'].append ('toc-entry')

            if (self.backlinks in ('entry', 'top')
                 and title.next_node (nodes.Referential) is None):
                if self.backlinks == 'entry':
                    title['refid'] = ref_id
                elif self.backlinks == 'top' and toc_id is not None:
                    title['refid'] = toc_id

            entries.append (list_item)

        if entries:
            contents = nodes.bullet_list ('', *entries, **{'classes': ['compact', 'toc-list'],
                                                           'enumtype': 'none',
                                                           'pageno_maxlen': self.maxlen})
            self.startnode.replace_self (contents)
        else:
            self.startnode.parent.remove (self.startnode)
            # self.startnode.parent.parent.remove (self.startnode.parent)


class FootnotesDirectiveTransform (docutils.transforms.Transform):
    """
    Collects footnotes into this section.

    The 'footnotes::' directive creates a section and puts this
    pending transform into it.  This transform either moves all
    footnotes into the generated section or removes this section from
    non-HTML formats.

    """

    # run this before contents transform. contents should not grab
    # this section's title because it can disappear.

    default_priority = 718 # before empty section remover

    def apply (self, **kwargs):
        pending = self.startnode
        section = pending.parent
        section.remove (pending)

        writer = self.document.transformer.components['writer']
        if writer.supports ('html'):
            group = mynodes.footnote_group ()
            section += group

            for footnote in self.document.traverse (nodes.footnote):
                footnote.parent.remove (footnote)
                group += footnote
        #else:
        #    section.parent.remove (section)


class EmptySectionRemover (docutils.transforms.Transform):
    """
    Removes this section if it contains only a title or less.
    Use for footnotes or loa sections.

    """

    # run this before contents transform. contents should not grab
    # this section before we make it disappear.

    default_priority = 719 # before contents transform

    def apply (self, **kwargs):
        pending = self.startnode
        section = pending.parent
        if not isinstance (section, nodes.section):
            return
        # print "\n\n", section
        index = section.index (pending)
        container = section[index + 1]
        # title = section[index - 1] if index > 0 else None
        title = section[0]
        if not isinstance (title, nodes.title):
            title = None
        section.remove (pending)

        if len (container) == 0:
            section.remove (container)
            if title is not None:
                section.remove (title)
            if len (section) == 0:
                section.parent.remove (section)


class PageNumberMoverTransform (docutils.transforms.Transform):
    """ Moves paragraphs that contain only one page number into the
    next paragraph. """

    default_priority = 721 # after contents transform

    def apply (self, **kwargs):
        for target in self.document.traverse (nodes.target):
            if isinstance (target.parent, nodes.paragraph) and len (target.parent) == 1:
                # move onto next appropriate node
                for next_node in target.traverse (nodes.TextElement, include_self = 0, ascend = 1):
                    if not isinstance (next_node, (nodes.Structural, nodes.Special)):
                        target.parent.parent.remove (target.parent)
                        next_node.insert (0, target)
                        break


class DropCapTransform (docutils.transforms.Transform):
    """ Inserts a dropcap into the following paragraph. """

    default_priority = 719 # run before Contents Transform

    def apply (self, **kwargs):
        iter_ = list (self.startnode.traverse (nodes.paragraph, siblings = 1))

        if len (iter_):
            para = iter_[0]
            iter_ = list (para.traverse (nodes.Text))
            details = self.startnode.details

            if len (iter_):
                textnode = iter_[0]
                charnode = spannode = restnode = None

                char = details['char']
                if not textnode.startswith (char):
                    error ("Dropcap: next paragraph doesn't start with: '%s'." % char)
                    return

                span = details.get ('span', '')
                if not textnode.startswith (span):
                    error ("Dropcap: next paragraph doesn't start with: '%s'." % span)
                    return
                if span and not span.startswith (char):
                    error ("Dropcap: span doesn't start with: '%s'." % char)
                    return
                if span == char:
                    span = ''

                if span:
                    # split into char/span/rest
                    restnode = nodes.Text (textnode.astext ()[len (span):])
                    spannode = nodes.inline ()
                    spannode.append (nodes.Text (textnode.astext ()[len (char):len (span)]))
                    spannode['classes'].append ('dropspan')
                else:
                    # split into char/rest
                    restnode = nodes.Text (textnode.astext ()[len (char):])
                    spannode = nodes.inline ('', '')
                    spannode['classes'].append ('dropspan')

                if 'image' in details:
                    charnode = nodes.image ()
                    charnode['uri'] = details['image']
                    charnode['alt'] = char
                    # debug ("Inserting image %s as dropcap." % uri)
                else:
                    charnode = nodes.inline ()
                    charnode.append (nodes.Text (char))
                    # debug ("Inserting char %s as dropcap." % char)

                charnode['classes'].append ('dropcap')
                charnode.attributes.update (details)

                para.replace (textnode, [charnode, spannode, restnode])

        self.startnode.parent.remove (self.startnode)


class InlineImageTransform (docutils.transforms.Transform):
    """Set class 'inline' or 'block' on an image according to actual
    usage so it can be used for styling.

    """

    default_priority = 730 # before StyleTransform

    def apply (self):
        for image in self.document.traverse (nodes.image):
            if isinstance (image.parent, nodes.TextElement):
                image['classes'] += ['inline']
            else:
                image['classes'] += ['block']


class StyleTransform (docutils.transforms.Transform):
    """
    Add classes to elements.

    Works in a way similar to CSS, though you can select only on
    element and class.

    Works on all following elements in the section it is used, if used
    before the first section works on the rest of the document.

    """

    default_priority = 731 # after default presentation

    def apply (self, **kwargs):
        pending  = self.startnode
        details  = pending.details

        if 'formats' in details:
            matched = False
            for f in details['formats'].split ():
                if f.startswith ('-'):
                    if not fnmatch.fnmatch (self.document.settings.format, f[1:]):
                        matched = True
                        break
                else:
                    if fnmatch.fnmatch (self.document.settings.format, f):
                        matched = True
                        break

            if not matched:
                pending.parent.remove (pending)
                return

        selector = details.get ('selector', '')

        if selector == 'document':
            # look at document node only
            node_list = [self.document]
        else:
            if 'titlehack' in details:
                # because titles start sections it is often impossible
                # to style titles because the pending node cannot be
                # placed before the title node.

                # traverse all children of parent
                node_list = list (pending.parent.traverse (
                    mynodes.node_selector (selector), include_self = 0))
            else:
                # traverse all following nodes and their children
                node_list = list (pending.traverse (
                    mynodes.node_selector (selector), siblings = 1, include_self = 0))

        # classes    = frozenset ([c for c in details.get ('class', []) if not c.startswith ('-')])
        # rmclasses  = frozenset ([c[1:] for c in details.get ('class', []) if c.startswith ('-')])

        classes = frozenset (details.get ('class', []))
        rmclasses = frozenset ()

        for n in node_list:
            n['classes'] = list ((set (n['classes']) | classes) - rmclasses)
            if details.get ('display', '').lower () == 'none':
                n.parent.remove (n)
                continue
            for a in ('align', 'width', 'float', 'hrules', 'vrules',
                      'aligns', 'vertical-aligns', 'tabularcolumns', 'widths'):
                if a in details:
                    n.setdefault (a, details[a]) # style on the element has priority
                    # n[a] = details[a]
            if 'before' in details:
                n[0:0] = [nodes.Text (details['before'])]
            if 'after' in details:
                index = len (n)
                n[index:index] = [nodes.Text (details['after'])]
            if pending.children:
                # replace element
                # print '****** replacing',  n.__class__.__name__
                n.parent.replace (n, [child.deepcopy () for child in pending.children])

        pending.parent.remove (pending)


###########
### 800 ###
###########


class FirstParagraphTransform (docutils.transforms.Transform):
    """
    Mark first paragraphs.

    With indented paragraphs, the first paragraph following a
    vertical space should not be indented. This transform tries
    to figure out which paragraphs should not be indented.

    Add the classes `pfirst` and `pnext` to paragraphs.

    """

    default_priority = 800 # late

    def apply (self, **kwargs):
        self.recurse (self.document, False)

    def recurse (self, parent, follows_paragraph):
        # follows_paragraph = False # flag: previous element is paragraph

        for node in parent:

            if isinstance (node, (nodes.paragraph)):
                node['classes'].append ('pnext' if follows_paragraph else 'pfirst')
                follows_paragraph = True
            elif isinstance (node, (nodes.title, nodes.subtitle)):
                # title may also be output as <html:p>
                node['classes'].append ('pfirst')
                follows_paragraph = False
            elif isinstance (node, (mynodes.page)):
                # explicit vertical space or page breaks
                follows_paragraph = False
            elif isinstance (node, (nodes.container, nodes.compound,
                                    nodes.Invisible, nodes.footnote, nodes.figure)):
                # invisible nodes are neutral, footnotes are not
                # output here so they are neutral too.  figures are
                # neutral because they can float away.  (also,
                # paragraphs in real books may contain block figures
                # so not indenting the following paragraph would look
                # ambiguous.)
                pass
            else:
                # everything else
                follows_paragraph = False

            if not isinstance (node, nodes.Text):
                self.recurse (node, follows_paragraph)



class BlockImageWrapper (docutils.transforms.Transform):
    """
    Wrap a block-level image into a figure.

    """

    default_priority = 801

    def apply (self):

        for image in self.document.traverse (nodes.image):

            # skip inline images
            # (See also: class `TextElement` in `docutils.nodes`.)
            if isinstance (image.parent, nodes.TextElement):
                continue

            # wrap all block images into figures
            if isinstance (image.parent, nodes.figure):
                figure = image.parent
            else:
                figure = nodes.figure ()
                figure['float'] = ('none', ) # do not float bare images
                figure['width'] = image.attributes.get ('width', 'image')
                figure['align'] = image.attributes.get ('align', 'center')
                image['width']  = '100%'
                image.replace_self (figure)
                figure.append (image)

            # set default width, align for block images only
            image.setdefault ('width', '100%')
            image.setdefault ('align', 'center')


class SetDefaults (docutils.transforms.Transform):
    """Set default attributes.

    Set default attributes to simplify writers.

    We need to set default attributes after the style directive
    because the style directive cannot distinguish between user-set
    attributes which must be kept and default attributes which must be
    overridden.

    Also, for simple tables and grid tables, we cannot set the default
    attributes in the directive parser because there is no directive.

    Image default attributes are set in BlockImageWrapper.

    """

    default_priority = 802

    def get_default_width (self, uri):
        """Calculate a sensible default width for images.

        Assume images are processed for a viewport 980px wide, the
        same as the iPhone browser assumes.

        """

        if (self.document.settings.get_image_size and
            six.callable (self.document.settings.get_image_size)):

            size = self.document.settings.get_image_size (uri)
            if size is not None:
                w = int (float (size[0]) / (980.0 * 0.8) * 100.0 + 0.5)
                width = "%d%%" % min (100, w)
                debug ('Got dimension of image: %s: %s' % (uri, width))
                return width

        warning ('Could not get dimension of image: %s' % uri)
        return '100%'


    def apply (self, **kwargs):

        def setdefault (node_type, l):
            for node in self.document.traverse (node_type):
                for name, default in l:
                    node.setdefault (name, default)

        # Image default attributes are set in BlockImageWrapper.

        setdefault (nodes.figure, (
            ('align',          'center'),
            ('float',          ('here', 'top', 'bottom', 'page')),
            ('width',          'image'),
        ))

        for image in self.document.traverse (nodes.image):
            if isinstance (image.parent, nodes.figure):
                figure = image.parent
                if figure['width'] == 'image':
                    figure['width'] = self.get_default_width (image['uri'])
                    figure['classes'].append ('auto-scaled')

        setdefault (nodes.table, (
            ('align',          'center'),
            ('float',          ('here', 'top', 'bottom', 'page')),
            ('hrules',         ('table', 'rows')),
            ('summary',        'no summary'),
            ('tabularcolumns', None),
            ('vrules',         ('none', )),
            ('width',          '100%'),
        ))
        setdefault (nodes.topic, (
            ('float',          ('here', )),
        ))
        setdefault (nodes.sidebar, (
            ('float',          ('here', )),
        ))


class AlignTransform (docutils.transforms.Transform):
    """
    Transforms align attribute into align-* class.

    """

    default_priority = 803 # after SetDefaults

    def apply (self):
        for body in self.document.traverse (
            lambda x: isinstance (x, (nodes.Body, nodes.Structural))):
            if 'align' in body:
                body['classes'].append ('align-%s' % body['align'])


class TextTransform (docutils.transforms.Transform):
    """
    Implements CSS text-transform.

    """

    default_priority = 895 # next to last

    smartquotes_map = {
        0x0027: '’',
        # 0x0022: '”',
        }

    re_quotes_thin_space = re.compile (r'([“„‟’])([‘‚‛”])')

    def apply (self, **kwargs):
        self.recurse (self.document, {})

    # FIXME this has been obsoleted by class inheritance

    def recurse (self, node, text_transform):
        if isinstance (node, nodes.Text):
            if len (text_transform) > 0:
                oldtext = text = node.astext ()
                if text_transform.get ('uppercase', False):
                    text = text.upper ()
                if text_transform.get ('smartquotes', False):
                    text = text.translate (self.smartquotes_map)
                    # insert thin space between quotes
                    text = self.re_quotes_thin_space.sub (r'\1 \2', text)
                if text != oldtext:
                    node.parent.replace (node, nodes.Text (text)) # cannot change text nodes
            return

        ntt = text_transform.copy ()
        classes = node['classes']

        if 'text-transform-uppercase' in classes:
            ntt['uppercase'] = True
        elif 'text-transform-smartquotes' in classes:
            ntt['smartquotes'] = True
        elif 'text-transform-none' in classes:
            ntt['uppercase'] = False
            ntt['smartquotes'] = False

        if 'white-space-pre-line' in classes:
            ntt['pre-line'] = True

        for child in node:
            self.recurse (child, ntt)


class CharsetTransform (docutils.transforms.Transform):
    """
    Translates text into smaller charset.

    This does not change the encoding, it just emulates all characters
    with characters from the smaller charset.

    """

    default_priority = 896

    def apply (self, **kwargs):
        if self.document.settings.encoding != 'utf-8':
            charset = self.document.settings.encoding
            del Unitame.unhandled_chars[:]

            for n in self.document.traverse (nodes.Text):
                text  = n.astext ()
                text2 = text.encode (charset, 'unitame').decode (charset)
                if text != text2:
                    n.parent.replace (n, nodes.Text (text2)) # cannot change text nodes

            if Unitame.unhandled_chars:
                error ("unitame: unhandled chars: %s" % ", ".join (set (Unitame.unhandled_chars)))


class TextNodeWrapper (docutils.transforms.Transform):
    """
    Wrap all naked Text nodes in inline nodes.

    Works in conjunction with the inheritance transform.

    """

    default_priority = 897 # before NodeTypeTransform

    def apply (self):
        for node in self.document.traverse (nodes.Text):

            # skip already wrapped nodes
            if isinstance (node.parent, nodes.Inline):
                continue

            node.parent.replace (node, nodes.inline ('', node.astext ()))


class NodeTypeTransform (docutils.transforms.Transform):
    """
    Determines the type of a node.

    The type can be one of 'empty', 'text', 'inline', 'simple', 'compound'.

    - An empty element contains no other elements nor text. Eg. image,
      transition.

    - A text element contains text.

    - An inline element contains other inline elements or text.

    - A simple element is a block element that contains text or inline
      elements but never contains other block elements.  It is the
      innermost block element.

    - A compound element is a block element that contains other simple
      or compound block elements but never contains inline or text
      elements.

    See: http://docutils.sourceforge.net/docs/ref/doctree.html#body-elements

    """

    # FIXME: there should be a class factory for nodes in docutils, so
    # you could add some type of identification function.

    default_priority = 898

    def apply (self, **kwargs):
        for node in self.document.traverse ():
            node.type = \
                'text'   if isinstance (node, nodes.Text) else (
                'inline' if isinstance (node, nodes.Inline) else (
                'simple' if isinstance (node, nodes.TextElement) else
                'empty'  if isinstance (node, (nodes.transition, nodes.image)) else
                'compound'))

            # this helps distinguish the use of elements than can be
            # both blocks or inline elements (eg. images)
            node.is_block = not node.parent or not isinstance (node.parent, (nodes.TextElement))


class InheritTransform (docutils.transforms.Transform):
    """
    Inheritance for docutil classes. Abstract base class.

    For convenience of use, you may specify a class on a block, that
    has an effect only on inline text. Eg. you may specify a class
    'italic' on a block to make the whole block italic.

    Inheritance is a way to get classes down from the (block) element
    where they are specified onto the (inline) element where they are
    applied.

    Override this class in the writer to gain finer control over which
    classes are inherited.

    """

    default_priority = 899 # last, you should not change the tree after this

    # FIXME: find a way to let the writer specify the classes
    # eg HTML => white-space-pre-line only

    apply_to_simple = frozenset ("""
    left center right justify
    noindent
    """.split ())

    apply_to_inline = frozenset ("""
    italics bold small-caps gesperrt normal antiqua monospaced
    xx-small x-small small medium large x-large xx-large
    smaller larger
    red green blue yellow white gray black
    """.split ())

    apply_to_text = frozenset ("""
    white-space-pre-line
    """.split ())

    # pass on if it applies to a contained class:
    #    compound > simple > inline > text

    pass_on = {
        'text':     frozenset (),
        'inline':   apply_to_text,
        'simple':   apply_to_inline | apply_to_text,
        'compound': apply_to_simple | apply_to_inline | apply_to_text,
        }

    def apply (self, **kwargs):
        self.recurse (None, self.document, set ())

    def recurse (self, parent, node, inherited_classes):
        classes_to_pass_on = self.pass_on[node.type]

        classes = set (node['classes']) | inherited_classes

        # remove classes with 'no-' prefix
        classes -= set ([c[3:] for c in classes if c.startswith ('no-')])

        # remove from node classes we passed on
        # this avoids problems with classes like 'smaller'
        node['classes'] = list (classes - classes_to_pass_on)

        passed_on_classes = classes & classes_to_pass_on

        for n in node.children:
            if isinstance (n, nodes.Text):
                n.attributes = {'classes': list (passed_on_classes) } # HACK! Text has no attributes
            else:
                self.recurse (node, n, passed_on_classes)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

from __future__ import unicode_literals

unicode_to_iso_8859_1 = {
    'Đ': 'D', # LATIN CAPITAL LETTER D WITH STROKE
    'đ': 'd', # LATIN SMALL LETTER D WITH STROKE
    'Ħ': 'H', # LATIN CAPITAL LETTER H WITH STROKE
    'ħ': 'h', # LATIN SMALL LETTER H WITH STROKE
    'Ŀ': 'L', # LATIN CAPITAL LETTER L WITH MIDDLE DOT
    'ŀ': 'l', # LATIN SMALL LETTER L WITH MIDDLE DOT
    'Ł': 'L', # LATIN CAPITAL LETTER L WITH STROKE
    'ł': 'l', # LATIN SMALL LETTER L WITH STROKE
    'ŉ': 'n', # LATIN SMALL LETTER N PRECEDED BY APOSTROPHE
    'Œ': 'OE', # LATIN CAPITAL LIGATURE OE
    'œ': 'oe', # LATIN SMALL LIGATURE OE
    'Ŧ': 'T', # LATIN CAPITAL LETTER T WITH STROKE
    'ŧ': 't', # LATIN SMALL LETTER T WITH STROKE
    'ƀ': 'b', # LATIN SMALL LETTER B WITH STROKE
    'Ɓ': 'B', # LATIN CAPITAL LETTER B WITH HOOK
    'Ƃ': 'B', # LATIN CAPITAL LETTER B WITH TOPBAR
    'ƃ': 'b', # LATIN SMALL LETTER B WITH TOPBAR
    'Ɔ': 'O', # LATIN CAPITAL LETTER OPEN O
    'Ƈ': 'C', # LATIN CAPITAL LETTER C WITH HOOK
    'ƈ': 'c', # LATIN SMALL LETTER C WITH HOOK
    'Ɗ': 'D', # LATIN CAPITAL LETTER D WITH HOOK
    'Ƌ': 'D', # LATIN CAPITAL LETTER D WITH TOPBAR
    'ƌ': 'd', # LATIN SMALL LETTER D WITH TOPBAR
    'Ƒ': 'F', # LATIN CAPITAL LETTER F WITH HOOK
    'ƒ': 'f', # LATIN SMALL LETTER F WITH HOOK
    'Ɠ': 'G', # LATIN CAPITAL LETTER G WITH HOOK
    'Ɨ': 'I', # LATIN CAPITAL LETTER I WITH STROKE
    'Ƙ': 'K', # LATIN CAPITAL LETTER K WITH HOOK
    'ƙ': 'k', # LATIN SMALL LETTER K WITH HOOK
    'ƚ': 'l', # LATIN SMALL LETTER L WITH BAR
    'Ɲ': 'N', # LATIN CAPITAL LETTER N WITH LEFT HOOK
    'ƞ': 'n', # LATIN SMALL LETTER N WITH LONG RIGHT LEG
    'Ɵ': 'O', # LATIN CAPITAL LETTER O WITH MIDDLE TILDE
    'Ƥ': 'P', # LATIN CAPITAL LETTER P WITH HOOK
    'ƥ': 'p', # LATIN SMALL LETTER P WITH HOOK
    'ƫ': 't', # LATIN SMALL LETTER T WITH PALATAL HOOK
    'Ƭ': 'T', # LATIN CAPITAL LETTER T WITH HOOK
    'ƭ': 't', # LATIN SMALL LETTER T WITH HOOK
    'Ʈ': 'T', # LATIN CAPITAL LETTER T WITH RETROFLEX HOOK
    'Ʋ': 'V', # LATIN CAPITAL LETTER V WITH HOOK
    'Ƴ': 'Y', # LATIN CAPITAL LETTER Y WITH HOOK
    'ƴ': 'y', # LATIN SMALL LETTER Y WITH HOOK
    'Ƶ': 'Z', # LATIN CAPITAL LETTER Z WITH STROKE
    'ƶ': 'z', # LATIN SMALL LETTER Z WITH STROKE
    'ǈ': 'L', # LATIN CAPITAL LETTER L WITH SMALL LETTER J
    'ǋ': 'N', # LATIN CAPITAL LETTER N WITH SMALL LETTER J
    'Ǣ': 'AE', # LATIN CAPITAL LETTER AE WITH MACRON
    'ǣ': 'ae', # LATIN SMALL LETTER AE WITH MACRON
    'Ǥ': 'G', # LATIN CAPITAL LETTER G WITH STROKE
    'ǥ': 'g', # LATIN SMALL LETTER G WITH STROKE
    'ǲ': 'D', # LATIN CAPITAL LETTER D WITH SMALL LETTER Z
    'Ǽ': 'AE', # LATIN CAPITAL LETTER AE WITH ACUTE
    'ǽ': 'ae', # LATIN SMALL LETTER AE WITH ACUTE
    'Ǿ': 'O', # LATIN CAPITAL LETTER O WITH STROKE AND ACUTE
    'ǿ': 'o', # LATIN SMALL LETTER O WITH STROKE AND ACUTE
    'Ƞ': 'N', # LATIN CAPITAL LETTER N WITH LONG RIGHT LEG
    'ȡ': 'd', # LATIN SMALL LETTER D WITH CURL
    'Ȥ': 'Z', # LATIN CAPITAL LETTER Z WITH HOOK
    'ȥ': 'z', # LATIN SMALL LETTER Z WITH HOOK
    'ȴ': 'l', # LATIN SMALL LETTER L WITH CURL
    'ȵ': 'n', # LATIN SMALL LETTER N WITH CURL
    'ȶ': 't', # LATIN SMALL LETTER T WITH CURL
    'ɓ': 'b', # LATIN SMALL LETTER B WITH HOOK
    'ɕ': 'c', # LATIN SMALL LETTER C WITH CURL
    'ɖ': 'd', # LATIN SMALL LETTER D WITH TAIL
    'ɗ': 'd', # LATIN SMALL LETTER D WITH HOOK
    'ɠ': 'g', # LATIN SMALL LETTER G WITH HOOK
    'ɦ': 'h', # LATIN SMALL LETTER H WITH HOOK
    'ɨ': 'i', # LATIN SMALL LETTER I WITH STROKE
    'ɫ': 'l', # LATIN SMALL LETTER L WITH MIDDLE TILDE
    'ɬ': 'l', # LATIN SMALL LETTER L WITH BELT
    'ɭ': 'l', # LATIN SMALL LETTER L WITH RETROFLEX HOOK
    'ɱ': 'm', # LATIN SMALL LETTER M WITH HOOK
    'ɲ': 'n', # LATIN SMALL LETTER N WITH LEFT HOOK
    'ɳ': 'n', # LATIN SMALL LETTER N WITH RETROFLEX HOOK
    'ɼ': 'r', # LATIN SMALL LETTER R WITH LONG LEG
    'ɽ': 'r', # LATIN SMALL LETTER R WITH TAIL
    'ɾ': 'r', # LATIN SMALL LETTER R WITH FISHHOOK
    'ʂ': 's', # LATIN SMALL LETTER S WITH HOOK
    'ʈ': 't', # LATIN SMALL LETTER T WITH RETROFLEX HOOK
    'ʉ': '', # LATIN SMALL LETTER U BAR
    'ʋ': 'v', # LATIN SMALL LETTER V WITH HOOK
    'ʐ': 'z', # LATIN SMALL LETTER Z WITH RETROFLEX HOOK
    'ʑ': 'z', # LATIN SMALL LETTER Z WITH CURL
    'ʜ': 'H', # LATIN LETTER SMALL CAPITAL H
    'ʝ': 'j', # LATIN SMALL LETTER J WITH CROSSED-TAIL
    'ʠ': 'q', # LATIN SMALL LETTER Q WITH HOOK
    'ʮ': 'h', # LATIN SMALL LETTER TURNED H WITH FISHHOOK
    'ʯ': 'h', # LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    'Ѝ': 'I', # CYRILLIC CAPITAL LETTER I WITH GRAVE
    'ѝ': 'i', # CYRILLIC SMALL LETTER I WITH GRAVE
    'Ӑ': 'A', # CYRILLIC CAPITAL LETTER A WITH BREVE
    'ӑ': 'a', # CYRILLIC SMALL LETTER A WITH BREVE
    'Ӓ': 'A', # CYRILLIC CAPITAL LETTER A WITH DIAERESIS
    'ӓ': 'a', # CYRILLIC SMALL LETTER A WITH DIAERESIS
    'Ӣ': 'I', # CYRILLIC CAPITAL LETTER I WITH MACRON
    'ӣ': 'i', # CYRILLIC SMALL LETTER I WITH MACRON
    'Ӥ': 'I', # CYRILLIC CAPITAL LETTER I WITH DIAERESIS
    'ӥ': 'i', # CYRILLIC SMALL LETTER I WITH DIAERESIS
    'Ӧ': 'O', # CYRILLIC CAPITAL LETTER O WITH DIAERESIS
    'ӧ': 'o', # CYRILLIC SMALL LETTER O WITH DIAERESIS
    'Ӭ': 'E', # CYRILLIC CAPITAL LETTER E WITH DIAERESIS
    'ӭ': 'e', # CYRILLIC SMALL LETTER E WITH DIAERESIS
    'Ӯ': '', # CYRILLIC CAPITAL LETTER U WITH MACRON
    'ӯ': '', # CYRILLIC SMALL LETTER U WITH MACRON
    'Ӱ': '', # CYRILLIC CAPITAL LETTER U WITH DIAERESIS
    'ӱ': '', # CYRILLIC SMALL LETTER U WITH DIAERESIS
    'Ӳ': '', # CYRILLIC CAPITAL LETTER U WITH DOUBLE ACUTE
    'ӳ': '', # CYRILLIC SMALL LETTER U WITH DOUBLE ACUTE
    'ẚ': 'a', # LATIN SMALL LETTER A WITH RIGHT HALF RING
    '‐': '-', # HYPHEN
    '–': '-', # EN DASH
    '—': '--', # EM DASH
    '‖': '||', # DOUBLE VERTICAL LINE
    '‗': '_', # DOUBLE LOW LINE
    '‘': '\'', # LEFT SINGLE QUOTATION MARK
    '’': '\'', # RIGHT SINGLE QUOTATION MARK
    '‚': '\'', # SINGLE LOW-9 QUOTATION MARK
    '‛': '\'', # SINGLE HIGH-REVERSED-9 QUOTATION MARK
    '“': '"', # LEFT DOUBLE QUOTATION MARK
    '”': '"', # RIGHT DOUBLE QUOTATION MARK
    '„': '"', # DOUBLE LOW-9 QUOTATION MARK
    '‟': '"', # DOUBLE HIGH-REVERSED-9 QUOTATION MARK
    '⁅': '[', # LEFT SQUARE BRACKET WITH QUILL
    '⁆': ']', # RIGHT SQUARE BRACKET WITH QUILL
}


iso_8859_1_to_ascii = {
    '¡': 'i', # INVERTED EXCLAMATION MARK
    '¢': 'c', # CENT SIGN
    '£': 'L', # POUND SIGN
    '¥': 'Y', # YEN SIGN
    '¦': '|', # BROKEN BAR
    '§': 'Sec.', # SECTION SIGN
    '¨': '"', # DIAERESIS
    '©': '(C)', # COPYRIGHT SIGN
    '«': '"', # LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    '­': '-', # SOFT HYPHEN
    '®': '(R)', # REGISTERED SIGN
    '¯': '-', # MACRON
    '°': ' deg.', # DEGREE SIGN
    '±': '+-', # PLUS-MINUS SIGN
    '²': '^2', # SUPERSCRIPT TWO
    '³': '^3', # SUPERSCRIPT THREE
    '´': '\'', # ACUTE ACCENT
    'µ': ' m', # MICRO SIGN
    '·': '.', # MIDDLE DOT
    '»': '"', # RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    '¼': '1/4', # VULGAR FRACTION ONE QUARTER
    '½': '1/2', # VULGAR FRACTION ONE HALF
    '¾': '3/4', # VULGAR FRACTION THREE QUARTERS
    '¿': '?', # INVERTED QUESTION MARK
    'Ä': 'Ae', # LATIN CAPITAL LETTER A WITH DIAERESIS
    'Æ': 'AE', # LATIN CAPITAL LETTER AE
    'Ð': 'Eth', # LATIN CAPITAL LETTER ETH
    'Ö': 'Oe', # LATIN CAPITAL LETTER O WITH DIAERESIS
    '×': 'x', # MULTIPLICATION SIGN
    'Ø': 'O', # LATIN CAPITAL LETTER O WITH STROKE
    'Ü': 'Ue', # LATIN CAPITAL LETTER U WITH DIAERESIS
    'ß': 'ss', # LATIN SMALL LETTER SHARP S
    'ä': 'ae', # LATIN SMALL LETTER A WITH DIAERESIS
    'æ': 'ae', # LATIN SMALL LETTER AE
    'ð': 'eth', # LATIN SMALL LETTER ETH
    # 'ñ': 'ny', # LATIN SMALL LETTER N WITH TILDE
    'ö': 'oe', # LATIN SMALL LETTER O WITH DIAERESIS
    '÷': '/', # DIVISION SIGN
    'ø': 'o', # LATIN SMALL LETTER O WITH STROKE
    'ü': 'ue', # LATIN SMALL LETTER U WITH DIAERESIS
}



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-
"""

utils.py

tools for manipulating xhtml
Copyright 2009 by Project Gutenberg

Distributable under the GNU General Public License Version 3 or newer.
"""

import libgutenberg.GutenbergGlobals as gg
from libgutenberg.GutenbergGlobals import xpath, NS
from libgutenberg.Logger import critical, debug, error, info, warning

def css_len(len_str):
    """ if an int, make px """
    try:
        return str(int(len_str)) + 'px'
    except ValueError:
        return len_str

def add_class(elem, classname):
    if 'class' in elem.attrib and elem.attrib['class']:
        vals = elem.attrib['class'].split()
    else:
        vals = []
    vals.append(classname)
    elem.set('class', ' '.join(vals))

def add_style(elem, style=''):
    if style:
        if 'style' in elem.attrib and elem.attrib['style']:
            prev_style = elem.attrib['style'].strip(' ;')
            style = f'{style.strip(" ;")};{prev_style};'
        elem.set('style', style)

def check_lang(elem, lang_att):
    three2two = {'ita': 'it', 'lat': 'la', 'heb': 'he', 'fra': 'fr', 'spa': 'es', 'deu': 'de'}
    lang_att = three2two.get(lang_att, lang_att)
    lang = elem.attrib[lang_att]
    lang_name = gg.language_map.get(lang, default=None)
    if lang_name:
        if NS.xml.lang in elem.attrib:
             del elem.attrib[NS.xml.lang]
        elem.attrib['lang'] = lang
        return True
    clean_lang = gg.language_map.inverse(lang, default=None)
    if not clean_lang:
        warning("invalid lang attribute %s", lang)
        del elem.attrib[lang_att]
        elem.attrib['data-invalid-lang'] = lang
    elif lang != clean_lang:
        elem.attrib['lang'] = clean_lang
        if NS.xml.lang in elem.attrib:
             del elem.attrib[NS.xml.lang]

def replace_elements(xhtml, deprecated):
    ''' replace a dictionary of deprecated elements with a new element or just delete it.
        return a set of replaced elements 
    '''
    deprecated_used = set()
    for tag in deprecated:
        for elem in xpath(xhtml, "//xhtml:" + tag):
            if deprecated[tag]:
                add_class(elem, 'xhtml_' + tag)
                elem.tag = getattr(NS.xhtml, deprecated[tag])
            else:
                elem.getparent().remove(meta)
            deprecated_used.add(tag)
    return deprecated_used



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""

nodes.py

Copyright 2011 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Added nodes for PG.

"""

from docutils import nodes

class page (nodes.Element, nodes.Special):
    """ Hold pagination commands.

    Like clearpage, vspace etc.

    """

class newline (nodes.Element):
    """ A line break.

    Outputs a hard line break if the node or one of its parents belong
    to the class 'white-space-pre-line'.  Else a space.

    """

class footnote_group (nodes.container):
    """ Hold a group of footnotes. """


class variable (nodes.Inline, nodes.TextElement):
    """ A placeholder that gets substituted with actual text before output.

    We do not use substitution refs because they are resolved way too
    early in the transformation stage to be of much use to us.

    """


class node_selector (object):
    """ Allows CSS-like selectors as condition function for nodes.traverse (). """

    def __init__ (self, selector):

        # allow selectors like [element][.class[.class[...]]][, selector[, selector]]

        self.matches = [] # list of 2-tuples

        for sel in selector.split (','):
            sel = sel.strip ()
            if '.' not in sel:
                sel += '.'
            element, classes = sel.split ('.', 1)
            classes = set (classes.split ('.')) if classes else set ()
            self.matches.append ( (getattr (nodes, element, nodes.Element), classes) )


    def __call__ (self, node):
        """ returns True if the node matches the selector. """

        for match in self.matches:
            if isinstance (node, match[0]) and match[1].issubset (node['classes']):
                return True

        return False



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
HTMLPackager.py

Copyright 2010 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Package a HTML file for PG.

"""

from ebookmaker.packagers import HTMLishPackager

TYPE = 'ww'
FORMATS = 'html.images'.split ()

class Packager (HTMLishPackager):
    """ Package a HTML file with its images. """
    pass



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

WrapperParser.py

Copyright 2020 by Eric Hellman

Distributable under the GNU General Public License Version 3 or newer.

"""
from xml.sax.saxutils import escape, quoteattr

import lxml

from copy import copy
from libgutenberg.Logger import info
from libgutenberg import GutenbergGlobals as gg
from ebookmaker.parsers import HTMLParserBase, IMAGE_WRAPPER

mediatypes = ()

class Parser(HTMLParserBase):

    def __init__(self, attribs):
        HTMLParserBase.__init__(self, copy(attribs))
        self.attribs.orig_mediatype = self.attribs.mediatype
        self.src = attribs.url
        self.attribs.url = self.wrapper_url(attribs.url)
        self.attribs.orig_url = self.attribs.url
        self.attribs.nonlinear = True
        if not self.attribs.title:
            self.attribs.title = 'linked image'
        self.xhtml = lxml.etree.fromstring(
            self.unicode_content(),
            lxml.html.XHTMLParser(),
            base_url=self.attribs.url
        )
        self.fp = True  # so writers won't skip it

        # mark the image for treatment as a linked image
        attribs.rel.add('linked_image')
        # set the referrer for the image to this wrapper
        attribs.referrer = self.attribs.url


    def unicode_content(self):
        """ wrapper page content """
        frag = ('#%s' % self.attribs.id) if self.attribs.id else ''
        backlink = '<br /><a href="%s%s" title="back" >back</a>' % (
            escape(self.attribs.referrer), frag)
        return IMAGE_WRAPPER.format(
            src=escape(self.src),
            title=quoteattr(self.attribs.title),
            backlink=backlink,
            wrapper_class='x-ebookmaker-wrapper',
            doctype=gg.XHTML_DOCTYPE)


    def wrapper_url(self, img_url):
        """ make the wrapper url. """
        if self.attribs.id:
            return '%s.%s.wrap.html' % (img_url, self.attribs.id)
        return img_url + '.wrap.html'


    def make_toc(self, xhtml):
        return []


    def iterlinks(self):
        """ only return the image """
        for iterlink in super(Parser, self).iterlinks():
            if iterlink[1].tag == gg.NS.xhtml.img:
                yield iterlink


#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
TxtWriter.py

Copyright 2009 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Build an UTF-8-encoded PG plain text file. This is just the plain text
version recoded into UTF-8.

"""

from __future__ import unicode_literals

import os
import subprocess
import sys

from libgutenberg.Logger import debug, info, warning, error
from libgutenberg.GutenbergGlobals import SkipOutputFormat

from ebookmaker import ParserFactory
from ebookmaker import writers
from ebookmaker.CommonCode import Options

options = Options()

# map some not-widely-supported characters to more common ones
u2u = {
    0x2010: '-',  # unicode HYPHEN to HYPHEN-MINUS. Many Windows fonts lack this.
    }

class Writer(writers.BaseWriter):
    """ Class to write PG plain text. """

    def groff(self, job, nroff, encoding='utf-8'):
        """ Process thru groff.

        Takes and returns unicode strings!

        """

        device = {'utf-8': 'utf8',
                  'iso-8859-1': 'latin1',
                  'us-ascii': 'ascii'}[encoding]

        nroff = nroff.encode(encoding)
        nrofffilename = os.path.join(
            os.path.abspath(job.outputdir),
            os.path.splitext(job.outputfile)[0] + '.nroff')

        # write nroff file for debugging
        if options.verbose >= 2:
            with open(nrofffilename, 'wb') as fp:
                fp.write(nroff)
        else:
            try:
                # remove debug files from previous runs
                os.remove(nrofffilename)
            except OSError:
                pass

        # call groff
        try:
            _groff = subprocess.Popen([options.config.GROFF,
                                       "-t",             # preprocess with tbl
                                       "-K", device,     # input encoding
                                       "-T", device],    # output device
                                      stdin=subprocess.PIPE,
                                      stdout=subprocess.PIPE,
                                      stderr=subprocess.PIPE)
        except OSError:
            error("TxtWriter: executable not found: %s" % options.config.GROFF)
            raise SkipOutputFormat

        (txt, stderr) = _groff.communicate(nroff)

        # pylint: disable=E1103
        for line in stderr.splitlines():
            line = line.decode(sys.stderr.encoding)
            line = line.strip()
            if 'error' in line:
                error("groff: %s" % line)
            elif 'warn' in line:
                if options.verbose >= 1:
                    warning("groff: %s" % line)

        txt = txt.decode(encoding)
        return txt.translate(u2u) # fix nroff idiosyncracies


    def build(self, job):
        """ Build TXT file. """

        filename = os.path.join(job.outputdir, job.outputfile)

        encoding = job.subtype.strip('.')

        info("Creating plain text file: %s from %s", filename, job.url)

        parser = ParserFactory.ParserFactory.create(job.url)

        # don't make txt file unless the source is txt of some encoding
        has_txt_source = 'text/plain' in str(parser.attribs.orig_mediatype)
        is_html_source = not has_txt_source and \
                         hasattr(parser, 'xhtml') and \
                         parser.xhtml is not None

        if hasattr(parser, 'rst2nroff'):
            data = self.groff(job, parser.rst2nroff(job, encoding), encoding)
        elif is_html_source:
            info("Plain text file %s aborted due to html input" % filename)
            return
        else:
            data = parser.unicode_content()

        data = data.encode('utf_8_sig' if encoding == 'utf-8' else encoding, 'unitame')

        self.write_with_crlf(filename, data)

        info("Done plain text file: %s" % filename)
